from .version import __version__

import os
motd = r'''      ___           ___           ___           ___
     /__/\         /__/\         /__/\         /  /\      Version {:<8}
     \  \:\       |  |::\       |  |::\       /  /::\ 
      \  \:\      |  |:|:\      |  |:|:\     /  /:/\:\    An open source
  _____\__\:\   __|__|:|\:\   __|__|:|\:\   /  /:/  \:\   project originally
 /__/::::::::\ /__/::::| \:\ /__/::::| \:\ /__/:/ \__\:\  founded by Joseph Suarez
 \  \:\~~\~~\/ \  \:\~~\__\/ \  \:\~~\__\/ \  \:\ /  /:/  and formalized at OpenAI
  \  \:\  ~~~   \  \:\        \  \:\        \  \:\  /:/
   \  \:\        \  \:\        \  \:\        \  \:\/:/    Now developed and
    \  \:\        \  \:\        \  \:\        \  \::/     maintained at MIT in
     \__\/         \__\/         \__\/         \__\/      Phillip Isola's lab '''.format(__version__)

from . import scripting
from .lib import material, spawn
from .overlay import Overlay, OverlayRegistry
from .io import action
from .io.stimulus import Serialized
from .io.action import Action
from .core import config, agent
from .core.agent import Agent
from .core.env import Env, Replay
from . import scripting, emulation, integrations
from .systems.achievement import Task
from .core.terrain import MapGenerator, Terrain

__all__ = ['Env', 'config', 'scripting', 'emulation', 'integrations', 'agent', 'Agent', 'MapGenerator', 'Terrain',
        'Serialized', 'action', 'Action', 'scripting', 'material', 'spawn',
        'Task', 'Overlay', 'OverlayRegistry', 'Replay']

try:
    import openskill
    from .lib.rating import OpenSkillRating
    __all__.append('OpenSkillRating')
except:
    print('Warning: OpenSkill not installed. Ignore if you do not need this feature')



import numpy as np
from pdb import set_trace as T

import nmmo
from nmmo.systems import ai, equipment, inventory
from nmmo.lib import material

from nmmo.systems.skill import Skills
from nmmo.systems.achievement import Diary
from nmmo.systems import combat
from nmmo.entity import entity

class Player(entity.Entity):
   def __init__(self, realm, pos, agent, color, pop):
      super().__init__(realm, pos, agent.iden, agent.policy, color, pop)
      self.agent  = agent
      self.pop    = pop

      # Scripted hooks
      self.target = None
      self.food   = None
      self.water  = None
      self.vision = 7

      # Logs
      self.buys                     = 0
      self.sells                    = 0 
      self.ration_consumed          = 0
      self.poultice_consumed        = 0
      self.ration_level_consumed    = 0
      self.poultice_level_consumed  = 0

      # Submodules
      self.skills = Skills(realm, self)

      self.diary  = None
      tasks = realm.config.TASKS
      if tasks:
          self.diary = Diary(tasks)

      self.dataframe.init(nmmo.Serialized.Entity, self.entID, self.pos)

   @property
   def serial(self):
      return self.population, self.entID

   @property
   def isPlayer(self) -> bool:
      return True

   @property
   def population(self):
      if __debug__:
          assert self.base.population.val == self.pop
      return self.pop

   @property
   def level(self) -> int:
       return combat.level(self.skills)

   def applyDamage(self, dmg, style):
      super().applyDamage(dmg, style)
      self.skills.applyDamage(dmg, style)
      
   def receiveDamage(self, source, dmg):
      if super().receiveDamage(source, dmg):
          return True
     
      if not self.config.ITEM_SYSTEM_ENABLED:
          return False

      for item in list(self.inventory._item_references):
          if not item.quantity.val:
              continue

          self.inventory.remove(item)

          if source.inventory.space:
              source.inventory.receive(item)

      if not super().receiveDamage(source, dmg):
         if source:
            source.history.playerKills += 1
         return 

      self.skills.receiveDamage(dmg)

   @property
   def equipment(self):
       return self.inventory.equipment

   def packet(self):
      data = super().packet()

      data['entID']     = self.entID
      data['annID']     = self.population

      data['base']      = self.base.packet()
      data['resource']  = self.resources.packet()
      data['skills']    = self.skills.packet()
      data['inventory'] = self.inventory.packet()

      return data
  
   def update(self, realm, actions):
      '''Post-action update. Do not include history'''
      super().update(realm, actions)

      # Spawsn battle royale style death fog
      # Starts at 0 damage on the specified config tick
      # Moves in from the edges by 1 damage per tile per tick
      # So after 10 ticks, you take 10 damage at the edge and 1 damage
      # 10 tiles in, 0 damage in farther
      # This means all agents will be force killed around 
      # MAP_CENTER / 2 + 100 ticks after spawning
      fog = self.config.PLAYER_DEATH_FOG
      if fog is not None and self.realm.tick >= fog:
          r, c = self.pos
          cent = self.config.MAP_BORDER + self.config.MAP_CENTER // 2

          # Distance from center of the map
          dist = max(abs(r - cent), abs(c - cent))

          # Safe final area
          if dist > self.config.PLAYER_DEATH_FOG_FINAL_SIZE:
              # Damage based on time and distance from center
              time_dmg = self.config.PLAYER_DEATH_FOG_SPEED * (self.realm.tick - fog + 1)
              dist_dmg = dist - self.config.MAP_CENTER // 2
              dmg = max(0, dist_dmg + time_dmg)
              self.receiveDamage(None, dmg)

      if not self.alive:
         return

      self.resources.update(realm, self, actions)
      self.skills.update(realm, self)

      if self.diary:
         self.diary.update(realm, self)



from pdb import set_trace as T
import numpy as np

from collections import defaultdict
import itertools

import gym

import nmmo
from nmmo.infrastructure import DataType

class SingleAgentEnv:
    def __init__(self, env, idx, max_idx):
        self.config = env.config
        self.env    = env
        self.idx    = idx
        self.last   = idx == max_idx 

    def reset(self):
        if not self.env.has_reset:
            self.obs = self.env.reset()

        return self.obs[self.idx]            

    def step(self, actions):
        if self.last:
            self.obs, self.rewards, self.dones, self.infos = self.env.step(actions)

        i = self.idx
        return self.obs[i], self.rewards[i], self.dones[i], self.infos[i]

def multiagent_to_singleagent(config):
    assert config.EMULATE_CONST_PLAYER_N, "Wrapper requires constant num agents"

    base_env = nmmo.Env(config)
    n = config.PLAYER_N

    return [SingleAgentEnv(base_env, i, n) for i in range(1, n+1)]
        
def pad_const_nent(config, dummy_ob, obs, rewards, dones, infos):
    for i in range(1, config.PLAYER_N+1):                               
        if i not in obs:                                                  
            obs[i]     = dummy_ob                                         
            rewards[i] = 0                                                 
            infos[i]   = {}
            dones[i]   = False

def const_horizon(dones):
    for agent in dones:
        dones[agent] = True

    return dones

def pack_atn_space(config):
   actions = defaultdict(dict)                                             
   for atn in sorted(nmmo.Action.edges(config)):
      for arg in sorted(atn.edges):
         actions[atn][arg] = arg.N(config)

   n = 0                                                                   
   flat_actions = {}                                                  
   for atn, args in actions.items():                                       
       ranges = [range(e) for e in args.values()]                          
       for vals in itertools.product(*ranges):                                       
          flat_actions[n] = {atn: {arg: val for arg, val in zip(args, vals)}}
          n += 1 

   return flat_actions

def pack_obs_space(observation):
   n = 0                                                                   
   #for entity, obs in observation.items():                                 
   for entity in observation:                                 
      obs = observation[entity]
      #for attr_name, attr_box in obs.items():                              
      for attr_name in obs:                              
         attr_box = obs[attr_name]
         n += np.prod(observation[entity][attr_name].shape)                
                                                                           
   return gym.spaces.Box(
           low=-2**20, high=2**20,
           shape=(int(n),), dtype=DataType.CONTINUOUS)


def batch_obs(config, obs):
    batched = {}
    for (entity_name,), entity in nmmo.io.stimulus.Serialized:
        if not entity.enabled(config):
            continue

        batched[entity_name] = {}
        for dtype in 'Continuous Discrete N'.split():
            attr_obs = [obs[k][entity_name][dtype] for k in obs]
            batched[entity_name][dtype] = np.stack(attr_obs, 0)

    return batched

def pack_obs(obs):
    packed = {}
    for key in obs:
        ary = []
        obs[key].items()
        for ent_name, ent_attrs in obs[key].items():
            for attr_name, attr in ent_attrs.items():
                ary.append(attr.ravel())
        packed[key] = np.concatenate(ary) 

    return packed 

def unpack_obs(config, packed_obs):
    obs, idx = {}, 0
    batch = len(packed_obs)
    for (entity_name,), entity in nmmo.io.stimulus.Serialized:
        if not entity.enabled(config):
            continue

        n_entity = entity.N(config)
        n_continuous, n_discrete = 0, 0
        obs[entity_name] = {}

        for attribute_name, attribute in entity:
            if attribute.CONTINUOUS:
                n_continuous += 1
            if attribute.DISCRETE:
                n_discrete += 1

        inc = int(n_entity * n_continuous)
        obs[entity_name]['Continuous'] = packed_obs[:, idx: idx + inc].reshape(batch, n_entity, n_continuous)
        idx += inc

        inc = int(n_entity * n_discrete)
        obs[entity_name]['Discrete'] = packed_obs[:, idx: idx + inc].reshape(batch, n_entity, n_discrete)
        idx += inc

        inc = 1
        obs[entity_name]['N'] = packed_obs[:, idx: idx + inc].reshape(batch, 1)
        idx += inc

    return obs



from pdb import set_trace as T
import logging
import abc

import numpy as np
from nmmo.io.stimulus import Serialized
from nmmo.systems import experience, combat, ai
from nmmo.lib import material

### Infrastructure ###
class SkillGroup:
   def __init__(self, realm, entity):
      self.config  = realm.config
      self.realm   = realm

      self.expCalc = experience.ExperienceCalculator()
      self.skills  = set()

   def update(self, realm, entity):
       for skill in self.skills:
           skill.update(realm, entity)

   def packet(self):
      data = {}
      for skill in self.skills:
         data[skill.__class__.__name__.lower()] = skill.packet()

      return data

class Skill:
   skillItems = abc.ABCMeta

   def __init__(self, realm, entity, skillGroup):
      self.config  = realm.config
      self.realm   = realm

      self.expCalc = skillGroup.expCalc
      self.exp     = 0

      skillGroup.skills.add(self)

   def packet(self):
      data = {}

      data['exp']   = self.exp
      data['level'] = self.level.val

      return data

   def add_xp(self, xp):
      level     = self.expCalc.levelAtExp(self.exp)
      self.exp += xp * self.config.PROGRESSION_BASE_XP_SCALE

      level = self.expCalc.levelAtExp(self.exp)
      self.level.update(int(level))

      if self.config.LOG_MILESTONES and self.realm.quill.milestone.log_max(f'Level_{self.__class__.__name__}', level) and self.config.LOG_VERBOSE:
         logging.info(f'PROGRESSION: Reached level {level} {self.__class__.__name__}')

   def setExpByLevel(self, level):
      self.exp = self.expCalc.expAtLevel(level)
      self.level.update(int(level))

### Skill Bases ###
class CombatSkill(Skill):
  def update(self, realm, entity):
     pass

class NonCombatSkill(Skill): pass


class HarvestSkill(NonCombatSkill):
    def processDrops(self, realm, entity, matl, dropTable):
        level = 1
        tool  = entity.equipment.held
        if type(tool) == matl.tool:
            level = tool.level.val

        #TODO: double-check drop table quantity
        for drop in dropTable.roll(realm, level):
            assert drop.level.val == level, 'Drop level does not match roll specification'

            if self.config.LOG_MILESTONES and realm.quill.milestone.log_max(f'Gather_{drop.__class__.__name__}', level) and self.config.LOG_VERBOSE:
                logging.info(f'PROFESSION: Gathered level {level} {drop.__class__.__name__} (level {self.level.val} {self.__class__.__name__})') 

            if entity.inventory.space:
                entity.inventory.receive(drop)

    def harvest(self, realm, entity, matl, deplete=True):
        r, c = entity.pos
        if realm.map.tiles[r, c].state != matl:
            return

        dropTable = realm.map.harvest(r, c, deplete)
        if dropTable:
            self.processDrops(realm, entity, matl, dropTable)
            return True

    def harvestAdjacent(self, realm, entity, matl, deplete=True):
        r, c      = entity.pos
        dropTable = None

        if realm.map.tiles[r-1, c].state == matl:
            dropTable = realm.map.harvest(r-1, c, deplete)
        if realm.map.tiles[r+1, c].state == matl:
            dropTable = realm.map.harvest(r+1, c, deplete)
        if realm.map.tiles[r, c-1].state == matl:
            dropTable = realm.map.harvest(r, c-1, deplete)
        if realm.map.tiles[r, c+1].state == matl:
            dropTable = realm.map.harvest(r, c+1, deplete)

        if dropTable:
            self.processDrops(realm, entity, matl, dropTable)
            return True

class AmmunitionSkill(HarvestSkill):
    def processDrops(self, realm, entity, matl, dropTable):
        super().processDrops(realm, entity, matl, dropTable)

        self.add_xp(self.config.PROGRESSION_AMMUNITION_XP_SCALE)


class ConsumableSkill(HarvestSkill):
    def processDrops(self, realm, entity, matl, dropTable):
        super().processDrops(realm, entity, matl, dropTable)

        self.add_xp(self.config.PROGRESSION_CONSUMABLE_XP_SCALE)


### Skill groups ###
class Basic(SkillGroup):
    def __init__(self, realm, entity):
        super().__init__(realm, entity)

        self.water = Water(realm, entity, self)
        self.food  = Food(realm, entity, self)

    @property
    def basicLevel(self):
        return 0.5 * (self.water.level
                + self.food.level)

class Harvest(SkillGroup):
    def __init__(self, realm, entity):
        super().__init__(realm, entity)

        self.fishing      = Fishing(realm, entity, self)
        self.herbalism    = Herbalism(realm, entity, self)
        self.prospecting  = Prospecting(realm, entity, self)
        self.carving      = Carving(realm, entity, self)
        self.alchemy      = Alchemy(realm, entity, self)

    @property
    def harvestLevel(self):
        return max(self.fishing.level,
                   self.herbalism.level,
                   self.prospecting.level,
                   self.carving.level,
                   self.alchemy.level) 

class Combat(SkillGroup):
   def __init__(self, realm, entity):
      super().__init__(realm, entity)

      self.melee = Melee(realm, entity, self)
      self.range = Range(realm, entity, self)
      self.mage  = Mage(realm, entity, self)

   def packet(self):
      data          = super().packet() 
      data['level'] = combat.level(self)

      return data

   @property
   def combatLevel(self):
      return max(self.melee.level,
                 self.range.level,
                 self.mage.level)

   def applyDamage(self, dmg, style):
      if not self.config.PROGRESSION_SYSTEM_ENABLED:
         return

      config = self.config
      skill  = self.__dict__[style]
      skill.add_xp(config.PROGRESSION_COMBAT_XP_SCALE)

   def receiveDamage(self, dmg):
      pass

class Skills(Basic, Harvest, Combat):
    pass

### Skills ###
class Melee(CombatSkill):
    def __init__(self, realm, ent, skillGroup):
        self.level = Serialized.Entity.Melee(ent.dataframe, ent.entID)
        super().__init__(realm, ent, skillGroup)

class Range(CombatSkill):
    def __init__(self, realm, ent, skillGroup):
        self.level = Serialized.Entity.Range(ent.dataframe, ent.entID)
        super().__init__(realm, ent, skillGroup)

class Mage(CombatSkill):
    def __init__(self, realm, ent, skillGroup):
        self.level = Serialized.Entity.Mage(ent.dataframe, ent.entID)
        super().__init__(realm, ent, skillGroup)

Melee.weakness = Mage
Range.weakness = Melee
Mage.weakness  = Range

### Individual Skills ###

class CombatSkill(Skill): pass

class Lvl:
    def __init__(self, val):
        self.val = val

    def update(self, val):
        self.val = val

class Water(HarvestSkill):
    def __init__(self, realm, entity, skillGroup):
        self.level = Lvl(1)
        super().__init__(realm, entity, skillGroup)

    def update(self, realm, entity):
        config = self.config
        if not config.RESOURCE_SYSTEM_ENABLED:
            return

        depletion = config.RESOURCE_DEPLETION_RATE
        water = entity.resources.water
        water.decrement(depletion)

        tiles = realm.map.tiles
        if not self.harvestAdjacent(realm, entity, material.Water, deplete=False):
            return

        restore = np.floor(config.RESOURCE_BASE
                         * config.RESOURCE_HARVEST_RESTORE_FRACTION)
        water.increment(restore)

class Food(HarvestSkill):
    def __init__(self, realm, entity, skillGroup):
        self.level = Lvl(1)
        super().__init__(realm, entity, skillGroup)

    def update(self, realm, entity):
        config = self.config
        if not config.RESOURCE_SYSTEM_ENABLED:
            return

        depletion = config.RESOURCE_DEPLETION_RATE
        food = entity.resources.food
        food.decrement(depletion)

        if not self.harvest(realm, entity, material.Forest):
            return

        restore = np.floor(config.RESOURCE_BASE
                         * config.RESOURCE_HARVEST_RESTORE_FRACTION)
        food.increment(restore)

class Fishing(ConsumableSkill):
    def __init__(self, realm, ent, skillGroup):
        self.level = Serialized.Entity.Fishing(ent.dataframe, ent.entID)
        super().__init__(realm, ent, skillGroup)

    def update(self, realm, entity):
        self.harvestAdjacent(realm, entity, material.Fish)

class Herbalism(ConsumableSkill):
    def __init__(self, realm, ent, skillGroup):
        self.level = Serialized.Entity.Herbalism(ent.dataframe, ent.entID)
        super().__init__(realm, ent, skillGroup)

    def update(self, realm, entity):
        self.harvest(realm, entity, material.Herb)

class Prospecting(AmmunitionSkill):
    def __init__(self, realm, ent, skillGroup):
        self.level = Serialized.Entity.Prospecting(ent.dataframe, ent.entID)
        super().__init__(realm, ent, skillGroup)

    def update(self, realm, entity):
        self.harvest(realm, entity, material.Ore)

class Carving(AmmunitionSkill):
    def __init__(self, realm, ent, skillGroup):
        self.level = Serialized.Entity.Carving(ent.dataframe, ent.entID)
        super().__init__(realm, ent, skillGroup)

    def update(self, realm, entity):
        self.harvest(realm, entity, material.Tree)

class Alchemy(AmmunitionSkill):
    def __init__(self, realm, ent, skillGroup):
        self.level = Serialized.Entity.Alchemy(ent.dataframe, ent.entID)
        super().__init__(realm, ent, skillGroup)

    def update(self, realm, entity):
        self.harvest(realm, entity, material.Crystal)



from pdb import set_trace as T

import nmmo
from nmmo import Env

def rllib_env_cls():
    try:
        from ray import rllib
    except ImportError:
        raise ImportError('Integrations depend on rllib. Install ray[rllib] and then retry')
    class RLlibEnv(Env, rllib.MultiAgentEnv):
        def __init__(self, config):
            self.config = config['config']
            self.config.EMULATE_CONST_HORIZON = True
            super().__init__(self.config)

        def render(self):
            #Patrch of RLlib dupe rendering bug
            if not self.config.RENDER:
                return

            super().render()

        def step(self, actions):
            obs, rewards, dones, infos = super().step(actions)

            population  = len(self.realm.players) == 0
            hit_horizon = self.realm.tick >= self.config.EMULATE_CONST_HORIZON

            dones['__all__'] = False
            if not self.config.RENDER and (hit_horizon or population):
                dones['__all__'] = True

            return obs, rewards, dones, infos

    return RLlibEnv

class SB3Env(Env):
    def __init__(self, config):
        config.EMULATE_FLAT_OBS      = True
        config.EMULATE_FLAT_ATN      = True
        config.EMULATE_CONST_PLAYER_N = True
        config.EMULATE_CONST_HORIZON = True

        super().__init__(config)

    def step(self, actions):
        assert type(actions) == dict

        obs, rewards, dones, infos = super().step(actions)

        if self.realm.tick >= self.config.HORIZON or len(self.realm.players) == 0:
            # Cheat logs into infos
            stats = self.terminal()
            stats = {**stats['Env'], **stats['Player'], **stats['Milestone'], **stats['Event']}

            infos[1]['logs'] = stats

        return obs, rewards, dones, infos 

class CleanRLEnv(SB3Env):
    def __init__(self, config):
        super().__init__(config)

def sb3_vec_envs(config_cls, num_envs, num_cpus):
    try:
        import supersuit as ss
    except ImportError:
        raise ImportError('SB3 integration depend on supersuit. Install and then retry')

    config = config_cls()
    env    = SB3Env(config)

    env = ss.pettingzoo_env_to_vec_env_v1(env)
    env.black_death = True #We provide our own black_death emulation
    env = ss.concat_vec_envs_v1(env, num_envs, num_cpus,
            base_class='stable_baselines3')

    return env

def cleanrl_vec_envs(config_classes, verbose=True):
    '''Creates a vector environment object from a list of configs.

    Each subenv points to a single agent, but many agents can share the same env.
    All envs must have the same observation and action space, but they can have
    different numbers of agents'''

    try:
        import supersuit as ss
    except ImportError:
        raise ImportError('CleanRL integration depend on supersuit. Install and then retry')

    def make_env_fn(config_cls):
        '''Wraps the make_env fn to add a a config argument'''
        def make_env():
            config = config_cls()
            env    = CleanRLEnv(config)

            env = ss.pettingzoo_env_to_vec_env_v1(env)
            env.black_death = True #We provide our own black_death emulation

            env = ss.concat_vec_envs_v1(env,
                    config.NUM_ENVS // config.PLAYER_N,
                    config.NUM_CPUS,
                    base_class='gym')

            env.single_observation_space = env.observation_space
            env.single_action_space      = env.action_space
            env.is_vector_env            = True

            return env
        return make_env

    dummy_env = None
    all_envs = [] 

    num_cpus   = 0
    num_envs   = 0
    num_agents = 0

    if type(config_classes) != list:
        config_classes = [config_classes]

    for idx, cls in enumerate(config_classes):
        assert isinstance(cls, type), 'config_cls must be a type (did ytou pass an instance?)'
        assert hasattr(cls, 'NUM_ENVS'), f'config class {cls} must define NUM_ENVS'
        assert hasattr(cls, 'NUM_CPUS'), f'config class {cls} must define NUM_CPUS'
        assert isinstance(cls, type), f'config class {cls} must be a type (did you pass an instance?)'

        if dummy_env is None:
            config    = cls()
            dummy_env = CleanRLEnv(config)

        #neural = [e == nmmo.Agent for e in cls.PLAYERS]
        #n_neural = sum(neural) / len(neural) * config.NUM_ENVS
        #assert int(n_neural) == n_neural, f'{sum(neural)} neural agents and {cls.PLAYER_N} classes'
        #n_neural = int(n_neural)
        
        envs = make_env_fn(cls)#, n_neural)
        all_envs.append(envs)

        # TODO: Find a cleaner way to specify env scale that enables multiple envs per CPU
        # without having to pass multiple configs
        num_cpus    += cls.NUM_CPUS
        num_envs    += cls.NUM_CPUS
        num_agents  += cls.NUM_CPUS * cls.PLAYER_N



    envs = ss.vector.ProcConcatVec(all_envs,
            dummy_env.observation_space(1),
            dummy_env.action_space(1),
            num_agents,
            dummy_env.metadata)
    envs.is_vector_env = True

    if verbose:
        print(f'nmmo.integrations.cleanrl_vec_envs created {num_envs} envs across {num_cpus} cores')

    return envs



from pdb import set_trace as T
import numpy as np
import os

import nmmo
from nmmo.lib import utils, material, spawn


class Template(metaclass=utils.StaticIterable):
   def __init__(self):
      self.data = {}
      cls       = type(self)

      #Set defaults from static properties
      for k, v in cls:
         self.set(k, v)

   def override(self, **kwargs):
      for k, v in kwargs.items():
         err = 'CLI argument: {} is not a Config property'.format(k)
         assert hasattr(self, k), err
         self.set(k, v)

   def set(self, k, v):
      if type(v) is not property:
         try:
            setattr(self, k, v)
         except:
            print('Cannot set attribute: {} to {}'.format(k, v))
            quit()
      self.data[k] = v

   def print(self):
      keyLen = 0
      for k in self.data.keys():
         keyLen = max(keyLen, len(k))

      print('Configuration')
      for k, v in self.data.items():
         print('   {:{}s}: {}'.format(k, keyLen, v))

   def items(self):
       return self.data.items()

   def __iter__(self):
       for k in self.data:
           yield k

   def keys(self):
       return self.data.keys()

   def values(self):
       return self.data.values()

def validate(config):
    err = 'config.Config is a base class. Use config.{Small, Medium Large}'''
    assert type(config) != Config, err

    if not config.TERRAIN_SYSTEM_ENABLED:
        err = 'Invalid Config: {} requires Terrain'
        assert not config.RESOURCE_SYSTEM_ENABLED, err.format('Resource')
        assert not config.PROFESSION_SYSTEM_ENABLED, err.format('Profession')
 
    if not config.COMBAT_SYSTEM_ENABLED:
        err = 'Invalid Config: {} requires Combat'
        assert not config.NPC_SYSTEM_ENABLED, err.format('NPC')
 
    if not config.ITEM_SYSTEM_ENABLED:
        err = 'Invalid Config: {} requires Inventory'
        assert not config.EQUIPMENT_SYSTEM_ENABLED, err.format('Equipment')
        assert not config.PROFESSION_SYSTEM_ENABLED, err.format('Profession')
        assert not config.EXCHANGE_SYSTEM_ENABLED, err.format('Exchange')


class Config(Template):
   '''An environment configuration object

   Global constants are defined as static class variables. You can override
   any Config variable using standard CLI syntax (e.g. --NENT=128). 

   The default config as of v1.5 uses 1024x1024 maps with up to 2048 agents
   and 1024 NPCs. It is suitable to time horizons of 8192+ steps. For smaller
   experiments, consider the SmallMaps config.
   
   Notes:
      We use Google Fire internally to replace standard manual argparse
      definitions for each Config property. This means you can subclass
      Config to add new static attributes -- CLI definitions will be
      generated automatically.
   '''

   def __init__(self):
      super().__init__()

      # TODO: Come up with a better way
      # to resolve mixin MRO conflicts
      if not hasattr(self, 'TERRAIN_SYSTEM_ENABLED'):
          self.TERRAIN_SYSTEM_ENABLED = False

      if not hasattr(self, 'RESOURCE_SYSTEM_ENABLED'):
          self.RESOURCE_SYSTEM_ENABLED = False

      if not hasattr(self, 'COMBAT_SYSTEM_ENABLED'):
          self.COMBAT_SYSTEM_ENABLED = False

      if not hasattr(self, 'NPC_SYSTEM_ENABLED'):
          self.NPC_SYSTEM_ENABLED = False

      if not hasattr(self, 'PROGRESSION_SYSTEM_ENABLED'):
          self.PROGRESSION_SYSTEM_ENABLED = False

      if not hasattr(self, 'ITEM_SYSTEM_ENABLED'):
          self.ITEM_SYSTEM_ENABLED = False

      if not hasattr(self, 'EQUIPMENT_SYSTEM_ENABLED'):
          self.EQUIPMENT_SYSTEM_ENABLED = False

      if not hasattr(self, 'PROFESSION_SYSTEM_ENABLED'):
          self.PROFESSION_SYSTEM_ENABLED = False

      if not hasattr(self, 'EXCHANGE_SYSTEM_ENABLED'):
          self.EXCHANGE_SYSTEM_ENABLED = False

      if not hasattr(self, 'COMMUNICATION_SYSTEM_ENABLED'):  
          self.COMMUNICATION_SYSTEM_ENABLED = False
 
      if __debug__:
         validate(self)

      deprecated_attrs = [
            'NENT', 'NPOP', 'AGENTS', 'NMAPS', 'FORCE_MAP_GENERATION', 'SPAWN']

      for attr in deprecated_attrs:
          assert not hasattr(self, attr), f'{attr} has been deprecated or renamed'


   ############################################################################
   ### Meta-Parameters
   def game_system_enabled(self, name) -> bool:
      return hasattr(self, name)

   def population_mapping_fn(self, idx) -> int:
      return idx % self.NPOP

   RENDER                       = False
   '''Flag used by render mode'''

   SAVE_REPLAY            = False
   '''Flag used to save replays'''

   PLAYERS                      = []
   '''Player classes from which to spawn'''

   TASKS                        = []
   '''Tasks for which to compute rewards'''

   ############################################################################
   ### Emulation Parameters
 
   EMULATE_FLAT_OBS       = False
   '''Emulate a flat observation space'''

   EMULATE_FLAT_ATN       = False
   '''Emulate a flat action space'''

   EMULATE_CONST_PLAYER_N = False
   '''Emulate a constant number of agents'''

   EMULATE_CONST_HORIZON  = False
   '''Emulate a constant HORIZON simulations steps'''


   ############################################################################
   ### Population Parameters                                                   
   LOG_VERBOSE                  = False
   '''Whether to log server messages or just stats'''

   LOG_ENV                      = False
   '''Whether to log env steps (expensive)'''

   LOG_MILESTONES               = True
   '''Whether to log server-firsts (semi-expensive)'''

   LOG_EVENTS                   = True
   '''Whether to log events (semi-expensive)'''

   LOG_FILE                     = None
   '''Where to write logs (defaults to console)'''

   PLAYERS                      = []
   '''Player classes from which to spawn'''

   TASKS                        = []
   '''Tasks for which to compute rewards'''


   ############################################################################
   ### Player Parameters                                                   
   PLAYER_N                     = None                  
   '''Maximum number of players spawnable in the environment'''

   PLAYER_N_OBS                 = 100
   '''Number of distinct agent observations'''

   @property
   def PLAYER_POLICIES(self):
      '''Number of player policies'''
      return len(self.PLAYERS)

   PLAYER_BASE_HEALTH           = 100
   '''Initial agent health'''

   PLAYER_VISION_RADIUS         = 7
   '''Number of tiles an agent can see in any direction'''

   @property
   def PLAYER_VISION_DIAMETER(self):
      '''Size of the square tile crop visible to an agent'''
      return 2*self.PLAYER_VISION_RADIUS + 1

   PLAYER_DEATH_FOG             = None
   '''How long before spawning death fog. None for no death fog'''

   PLAYER_DEATH_FOG_SPEED       = 1
   '''Number of tiles per tick that the fog moves in'''

   PLAYER_DEATH_FOG_FINAL_SIZE  = 8
   '''Number of tiles from the center that the fog stops'''

   RESPAWN = False

   PLAYER_LOADER                = spawn.SequentialLoader
   '''Agent loader class specifying spawn sampling'''

   PLAYER_SPAWN_ATTEMPTS        = None
   '''Number of player spawn attempts per tick

   Note that the env will attempt to spawn agents until success
   if the current population size is zero.'''

   PLAYER_SPAWN_TEAMMATE_DISTANCE = 1
   '''Buffer tiles between teammates at spawn'''
   
   @property
   def PLAYER_SPAWN_FUNCTION(self):
      return spawn.spawn_concurrent

   @property
   def PLAYER_TEAM_SIZE(self):
      if __debug__:
         assert not self.PLAYER_N % len(self.PLAYERS)
      return self.PLAYER_N // len(self.PLAYERS)

   ############################################################################
   ### Map Parameters
   MAP_N                        = 1
   '''Number of maps to generate'''

   MAP_N_TILE                   = len(material.All.materials)
   '''Number of distinct terrain tile types'''

   @property
   def MAP_N_OBS(self):
      '''Number of distinct tile observations'''
      return int(self.PLAYER_VISION_DIAMETER ** 2)

   MAP_CENTER                   = None
   '''Size of each map (number of tiles along each side)'''

   MAP_BORDER                   = 16
   '''Number of lava border tiles surrounding each side of the map'''

   @property
   def MAP_SIZE(self):
      return int(self.MAP_CENTER + 2*self.MAP_BORDER)

   MAP_GENERATOR                = None
   '''Specifies a user map generator. Uses default generator if unspecified.'''

   MAP_FORCE_GENERATION         = True
   '''Whether to regenerate and overwrite existing maps'''

   MAP_GENERATE_PREVIEWS        = False
   '''Whether map generation should also save .png previews (slow + large file size)'''

   MAP_PREVIEW_DOWNSCALE        = 1
   '''Downscaling factor for png previews'''


   ############################################################################
   ### Path Parameters
   PATH_ROOT                = os.path.dirname(nmmo.__file__)
   '''Global repository directory'''

   PATH_CWD                 = os.getcwd()
   '''Working directory'''

   PATH_RESOURCE            = os.path.join(PATH_ROOT, 'resource')
   '''Resource directory'''

   PATH_TILE                = os.path.join(PATH_RESOURCE, '{}.png')
   '''Tile path -- format me with tile name'''

   PATH_MAPS                = None
   '''Generated map directory'''

   PATH_MAP_SUFFIX          = 'map{}/map.npy'
   '''Map file name'''

   PATH_MAP_SUFFIX          = 'map{}/map.npy'
   '''Map file name'''


############################################################################
### Game Systems (Static Mixins)
class Terrain:
   '''Terrain Game System'''

   TERRAIN_SYSTEM_ENABLED       = True
   '''Game system flag'''

   TERRAIN_FLIP_SEED            = False
   '''Whether to negate the seed used for generation (useful for unique heldout maps)'''

   TERRAIN_FREQUENCY            = -3
   '''Base noise frequency range (log2 space)'''

   TERRAIN_FREQUENCY_OFFSET     = 7
   '''Noise frequency octave offset (log2 space)'''

   TERRAIN_LOG_INTERPOLATE_MIN  = -2 
   '''Minimum interpolation log-strength for noise frequencies'''

   TERRAIN_LOG_INTERPOLATE_MAX  = 0
   '''Maximum interpolation log-strength for noise frequencies'''

   TERRAIN_TILES_PER_OCTAVE     = 8
   '''Number of octaves sampled from log2 spaced TERRAIN_FREQUENCY range'''

   TERRAIN_LAVA                 = 0.0
   '''Noise threshold for lava generation'''

   TERRAIN_WATER                = 0.30
   '''Noise threshold for water generation'''

   TERRAIN_GRASS                = 0.70
   '''Noise threshold for grass'''

   TERRAIN_FOREST               = 0.85
   '''Noise threshold for forest'''


class Resource:
   '''Resource Game System'''

   RESOURCE_SYSTEM_ENABLED             = True
   '''Game system flag'''

   RESOURCE_BASE                       = 100
   '''Initial level and capacity for Hunting + Fishing resource skills'''

   RESOURCE_DEPLETION_RATE             = 5
   '''Depletion rate for food and water'''

   RESOURCE_STARVATION_RATE            = 10
   '''Damage per tick without food'''

   RESOURCE_DEHYDRATION_RATE           = 10
   '''Damage per tick without water'''

   RESOURCE_FOREST_CAPACITY            = 1
   '''Maximum number of harvests before a forest tile decays'''

   RESOURCE_FOREST_RESPAWN             = 0.025
   '''Probability that a harvested forest tile will regenerate each tick'''

   RESOURCE_HARVEST_RESTORE_FRACTION   = 1.0
   '''Fraction of maximum capacity restored upon collecting a resource'''

   RESOURCE_HEALTH_REGEN_THRESHOLD     = 0.5
   '''Fraction of maximum resource capacity required to regen health'''

   RESOURCE_HEALTH_RESTORE_FRACTION    = 0.1
   '''Fraction of health restored per tick when above half food+water'''


class Combat:
   '''Combat Game System'''

   COMBAT_SYSTEM_ENABLED              = True
   '''Game system flag'''

   COMBAT_FRIENDLY_FIRE               = True
   '''Whether agents with the same population index can hit each other'''

   COMBAT_SPAWN_IMMUNITY              = 20
   '''Agents older than this many ticks cannot attack agents younger than this many ticks'''

   COMBAT_WEAKNESS_MULTIPLIER         = 1.5
   '''Multiplier for super-effective attacks'''

   def COMBAT_DAMAGE_FORMULA(self, offense, defense, multiplier):
       '''Damage formula'''
       return int(multiplier * (offense * (15 / (15 + defense))))

   COMBAT_MELEE_DAMAGE                = 30
   '''Melee attack damage'''

   COMBAT_MELEE_REACH                 = 3
   '''Reach of attacks using the Melee skill'''

   COMBAT_RANGE_DAMAGE                = 30
   '''Range attack damage'''

   COMBAT_RANGE_REACH                 = 3
   '''Reach of attacks using the Range skill'''

   COMBAT_MAGE_DAMAGE                 = 30
   '''Mage attack damage'''

   COMBAT_MAGE_REACH                  = 3
   '''Reach of attacks using the Mage skill'''


class Progression:
   '''Progression Game System'''

   PROGRESSION_SYSTEM_ENABLED        = True
   '''Game system flag'''

   PROGRESSION_BASE_XP_SCALE         = 1
   '''Base XP awarded for each skill usage -- multiplied by skill level'''

   PROGRESSION_COMBAT_XP_SCALE       = 1
   '''Multiplier on top of XP_SCALE for Melee, Range, and Mage'''

   PROGRESSION_AMMUNITION_XP_SCALE   = 1
   '''Multiplier on top of XP_SCALE for Prospecting, Carving, and Alchemy'''

   PROGRESSION_CONSUMABLE_XP_SCALE   = 5
   '''Multiplier on top of XP_SCALE for Fishing and Herbalism'''

   PROGRESSION_LEVEL_MAX             = 10
   '''Max skill level'''

   PROGRESSION_MELEE_BASE_DAMAGE     = 0
   '''Base Melee attack damage'''

   PROGRESSION_MELEE_LEVEL_DAMAGE    = 5
   '''Bonus Melee attack damage per level'''

   PROGRESSION_RANGE_BASE_DAMAGE     = 0
   '''Base Range attack damage'''

   PROGRESSION_RANGE_LEVEL_DAMAGE    = 5
   '''Bonus Range attack damage per level'''

   PROGRESSION_MAGE_BASE_DAMAGE      = 0
   '''Base Mage attack damage '''

   PROGRESSION_MAGE_LEVEL_DAMAGE     = 5
   '''Bonus Mage attack damage per level'''

   PROGRESSION_BASE_DEFENSE          = 0
   '''Base defense'''

   PROGRESSION_LEVEL_DEFENSE         = 5
   '''Bonus defense per level'''


class NPC:
   '''NPC Game System'''

   NPC_SYSTEM_ENABLED                  = True
   '''Game system flag'''

   NPC_N                               = None
   '''Maximum number of NPCs spawnable in the environment'''

   NPC_SPAWN_ATTEMPTS                  = 25
   '''Number of NPC spawn attempts per tick'''

   NPC_SPAWN_AGGRESSIVE                = 0.80
   '''Percentage distance threshold from spawn for aggressive NPCs'''

   NPC_SPAWN_NEUTRAL                   = 0.50
   '''Percentage distance threshold from spawn for neutral NPCs'''

   NPC_SPAWN_PASSIVE                   = 0.00
   '''Percentage distance threshold from spawn for passive NPCs'''
   
   NPC_LEVEL_MIN                       = 1
   '''Minimum NPC level'''

   NPC_LEVEL_MAX                       = 10
   '''Maximum NPC level'''

   NPC_BASE_DEFENSE                    = 0
   '''Base NPC defense'''

   NPC_LEVEL_DEFENSE                   = 30
   '''Bonus NPC defense per level'''

   NPC_BASE_DAMAGE                     = 15
   '''Base NPC damage'''

   NPC_LEVEL_DAMAGE                    = 30
   '''Bonus NPC damage per level'''


class Item:
   '''Inventory Game System'''

   ITEM_SYSTEM_ENABLED                 = True
   '''Game system flag'''

   ITEM_N                              = 17
   '''Number of unique base item classes'''

   ITEM_INVENTORY_CAPACITY             = 12
   '''Number of inventory spaces'''

   @property
   def ITEM_N_OBS(self):
       '''Number of distinct item observations'''
       return self.ITEM_N * self.NPC_LEVEL_MAX
       #return self.INVENTORY_CAPACITY


class Equipment:
   '''Equipment Game System'''

   EQUIPMENT_SYSTEM_ENABLED             = True
   '''Game system flag'''

   WEAPON_DROP_PROB = 0.025
   '''Chance of getting a weapon while harvesting ammunition'''

   EQUIPMENT_WEAPON_BASE_DAMAGE         = 15
   '''Base weapon damage'''

   EQUIPMENT_WEAPON_LEVEL_DAMAGE        = 15
   '''Added weapon damage per level'''

   EQUIPMENT_AMMUNITION_BASE_DAMAGE     = 15
   '''Base ammunition damage'''

   EQUIPMENT_AMMUNITION_LEVEL_DAMAGE    = 15
   '''Added ammunition damage per level'''

   EQUIPMENT_TOOL_BASE_DEFENSE          = 30
   '''Base tool defense'''

   EQUIPMENT_TOOL_LEVEL_DEFENSE         = 0
   '''Added tool defense per level'''

   EQUIPMENT_ARMOR_BASE_DEFENSE         = 0
   '''Base armor defense'''

   EQUIPMENT_ARMOR_LEVEL_DEFENSE        = 10
   '''Base equipment defense'''


class Profession:
   '''Profession Game System'''

   PROFESSION_SYSTEM_ENABLED           = True
   '''Game system flag'''

   PROFESSION_TREE_CAPACITY            = 1
   '''Maximum number of harvests before a tree tile decays'''

   PROFESSION_TREE_RESPAWN             = 0.105
   '''Probability that a harvested tree tile will regenerate each tick'''

   PROFESSION_ORE_CAPACITY             = 1
   '''Maximum number of harvests before an ore tile decays'''

   PROFESSION_ORE_RESPAWN              = 0.10
   '''Probability that a harvested ore tile will regenerate each tick'''

   PROFESSION_CRYSTAL_CAPACITY         = 1
   '''Maximum number of harvests before a crystal tile decays'''

   PROFESSION_CRYSTAL_RESPAWN          = 0.10
   '''Probability that a harvested crystal tile will regenerate each tick'''

   PROFESSION_HERB_CAPACITY            = 1
   '''Maximum number of harvests before an herb tile decays'''

   PROFESSION_HERB_RESPAWN             = 0.01
   '''Probability that a harvested herb tile will regenerate each tick'''

   PROFESSION_FISH_CAPACITY            = 1
   '''Maximum number of harvests before a fish tile decays'''

   PROFESSION_FISH_RESPAWN             = 0.01
   '''Probability that a harvested fish tile will regenerate each tick'''

   @staticmethod
   def PROFESSION_CONSUMABLE_RESTORE(level):
       return 50 + 5*level


class Exchange:
   '''Exchange Game System'''

   EXCHANGE_SYSTEM_ENABLED             = True
   '''Game system flag'''

   @property
   def EXCHANGE_N_OBS(self):
       '''Number of distinct item observations'''
       return self.ITEM_N * self.NPC_LEVEL_MAX

class Communication:
   '''Exchange Game System'''

   COMMUNICATION_SYSTEM_ENABLED             = True
   '''Game system flag'''

   @property
   def COMMUNICATION_NUM_TOKENS(self):
       '''Number of distinct item observations'''
       return self.ITEM_N * self.NPC_LEVEL_MAX


class AllGameSystems(Terrain, Resource, Combat, NPC, Progression, Item, Equipment, Profession, Exchange, Communication): pass


############################################################################
### Config presets
class Small(Config):
   '''A small config for debugging and experiments with an expensive outer loop'''

   PATH_MAPS                    = 'maps/small' 

   PLAYER_N                     = 64
   PLAYER_SPAWN_ATTEMPTS        = 1

   MAP_PREVIEW_DOWNSCALE        = 4
   MAP_CENTER                   = 32

   TERRAIN_LOG_INTERPOLATE_MIN  = 0

   NPC_N                        = 32
   NPC_LEVEL_MAX                = 5
   NPC_LEVEL_SPREAD             = 1

   PROGRESSION_SPAWN_CLUSTERS   = 4
   PROGRESSION_SPAWN_UNIFORMS   = 16

   HORIZON                      = 128


class Medium(Config):
   '''A medium config suitable for most academic-scale research'''

   PATH_MAPS                    = 'maps/medium' 

   PLAYER_N                     = 256
   PLAYER_SPAWN_ATTEMPTS        = 2

   MAP_PREVIEW_DOWNSCALE        = 16
   MAP_CENTER                   = 128

   NPC_N                        = 128
   NPC_LEVEL_MAX                = 10
   NPC_LEVEL_SPREAD             = 1

   PROGRESSION_SPAWN_CLUSTERS   = 64
   PROGRESSION_SPAWN_UNIFORMS   = 256

   HORIZON                      = 1024


class Large(Config):
   '''A large config suitable for large-scale research or fast models'''

   PATH_MAPS                    = 'maps/large' 

   PLAYER_N                     = 2048
   PLAYER_SPAWN_ATTEMPTS        = 16

   MAP_PREVIEW_DOWNSCALE        = 64
   MAP_CENTER                   = 1024

   NPC_N                        = 1024
   NPC_LEVEL_MAX                = 15
   NPC_LEVEL_SPREAD             = 3

   PROGRESSION_SPAWN_CLUSTERS   = 1024
   PROGRESSION_SPAWN_UNIFORMS   = 4096

   HORIZON                 = 8192


class Default(Medium, AllGameSystems): pass



from pdb import set_trace as T
import numpy as np

import inspect
import logging

from nmmo.systems import item as Item
from nmmo.systems import skill as Skill

class Equipment:
   def __init__(self, realm):
      self.hat         = None
      self.top         = None
      self.bottom      = None

      self.held        = None
      self.ammunition  = None

   def total(self, lambda_getter):
      items = [lambda_getter(e).val for e in self]
      if not items:
          return 0
      return sum(items)

   def __iter__(self):
      for item in [self.hat, self.top, self.bottom, self.held, self.ammunition]:
         if item is not None:
            yield item

   def conditional_packet(self, packet, item_name, item):
      if item:
         packet[item_name] = item.packet

   @property
   def item_level(self):
       return self.total(lambda e: e.level)

   @property
   def melee_attack(self):
       return self.total(lambda e: e.melee_attack)

   @property
   def range_attack(self):
       return self.total(lambda e: e.range_attack)

   @property
   def mage_attack(self):
       return self.total(lambda e: e.mage_attack)

   @property
   def melee_defense(self):
       return self.total(lambda e: e.melee_defense)

   @property
   def range_defense(self):
       return self.total(lambda e: e.range_defense)

   @property
   def mage_defense(self):
       return self.total(lambda e: e.mage_defense)

   @property
   def packet(self):
      packet = {}

      self.conditional_packet(packet, 'hat',        self.hat)
      self.conditional_packet(packet, 'top',        self.top)
      self.conditional_packet(packet, 'bottom',     self.bottom)
      self.conditional_packet(packet, 'held',       self.held)
      self.conditional_packet(packet, 'ammunition', self.ammunition)

      packet['item_level']    = self.item_level

      packet['melee_attack']  = self.melee_attack
      packet['range_attack']  = self.range_attack
      packet['mage_attack']   = self.mage_attack
      packet['melee_defense'] = self.melee_defense
      packet['range_defense'] = self.range_defense
      packet['mage_defense']  = self.mage_defense

      return packet


class Inventory:
   def __init__(self, realm, entity):
      config           = realm.config
      self.realm       = realm
      self.entity      = entity
      self.config      = config

      self.equipment   = Equipment(realm)

      if not config.ITEM_SYSTEM_ENABLED:
          return

      self.capacity         = config.ITEM_INVENTORY_CAPACITY
      self.gold             = Item.Gold(realm)

      self._item_stacks     = {self.gold.signature: self.gold}
      self._item_references = {self.gold}

   @property
   def space(self):
      return self.capacity - len(self._item_references)

   @property
   def dataframeKeys(self):
      return [e.instanceID for e in self._item_references]

   def packet(self):
      item_packet = []
      if self.config.ITEM_SYSTEM_ENABLED:
          item_packet = [e.packet for e in self._item_references]

      return {
            'items':     item_packet,
            'equipment': self.equipment.packet}

   def __iter__(self):
      for item in self._item_references:
         yield item

   def receive(self, item):
      assert isinstance(item, Item.Item), f'{item} received is not an Item instance'
      assert item not in self._item_references, f'{item} object received already in inventory'
      assert not item.equipped.val, f'Received equipped item {item}'
      assert self.space, f'Out of space for {item}'
      assert item.quantity.val, f'Received empty item {item}'

      config = self.config
      if config.LOG_MILESTONES and self.realm.quill.milestone.log_max(f'Receive_{item.__class__.__name__}', item.level.val) and config.LOG_VERBOSE:
          logging.info(f'INVENTORY: Received level {item.level.val} {item.__class__.__name__}')

      if isinstance(item, Item.Stack):
          signature = item.signature
          if signature in self._item_stacks:
              stack = self._item_stacks[signature]
              assert item.level.val == stack.level.val, f'{item} stack level mismatch'
              stack.quantity += item.quantity.val

              if config.LOG_MILESTONES and isinstance(item, Item.Gold) and self.realm.quill.milestone.log_max(f'Wealth', self.gold.quantity.val) and config.LOG_VERBOSE:
                  logging.info(f'EXCHANGE: Total wealth {self.gold.quantity.val} gold')
              
              return

          self._item_stacks[signature] = item

      self._item_references.add(item)

   def remove(self, item, quantity=None):
      assert isinstance(item, Item.Item), f'{item} received is not an Item instance'
      assert item in self._item_references, f'No item {item} to remove'

      if item.equipped.val:
          item.use(self.entity)

      assert not item.equipped.val, f'Removing {item} while equipped'

      if isinstance(item, Item.Stack):
         signature = item.signature 

         assert item.signature in self._item_stacks, f'{item} stack to remove not in inventory'
         stack = self._item_stacks[signature]

         if quantity is None or stack.quantity.val == quantity:
            self._item_references.remove(stack)
            del self._item_stacks[signature]
            return

         assert 0 < quantity <= stack.quantity.val, f'Invalid remove {quantity} x {item} ({stack.quantity.val} available)'
         stack.quantity.val -= quantity

         return

      self._item_references.remove(item)



from pdb import set_trace as T
from collections import defaultdict
from nmmo.lib import material
from copy import deepcopy
import os

import logging
import numpy as np
import json, pickle
import time

from nmmo.lib import utils

class Logger:
    def __init__(self):
        self.stats = defaultdict(list)

    def log(self, key, val):
        try:
            int_val = int(val)
        except TypeError as e:
            print(f'{val} must be int or float')
            raise e
        self.stats[key].append(val)
        return True

class MilestoneLogger(Logger):
    def __init__(self, log_file):
        super().__init__()
        logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO, filename=log_file, filemode='w')

    def log_min(self, key, val):
        if key in self.stats and val >= self.stats[key][-1]:
            return False

        self.log(key, val)
        return True

    def log_max(self, key, val):
        if key in self.stats and val <= self.stats[key][-1]:
            return False

        self.log(key, val)
        return True

class Quill:
   def __init__(self, config):
      self.config = config

      self.env    = Logger()
      self.player = Logger()
      self.event  = Logger()

      self.shared = {}

      if config.LOG_MILESTONES:
          self.milestone = MilestoneLogger(config.LOG_FILE)

   def register(self, key, fn):
       assert key not in self.shared, f'Log key {key} already exists'
       self.shared[key] = fn

   def log_env(self, key, val):
      self.env.log(key, val)

   def log_player(self, key, val):
      self.player.log(key, val)

   @property
   def packet(self):
      packet = {'Env':    self.env.stats,
              'Player': self.player.stats}

      if self.config.LOG_EVENTS:
          packet['Event'] = self.event.stats
      else:
          packet['Event'] = 'Unavailable: config.LOG_EVENTS = False'

      if self.config.LOG_MILESTONES:
          packet['Milestone'] = self.event.stats
      else:
          packet['Milestone'] = 'Unavailable: config.LOG_MILESTONES = False'

      return packet

#Log wrapper and benchmarker
class Benchmarker:
   def __init__(self, logdir):
      self.benchmarks = {}

   def wrap(self, func):
      self.benchmarks[func] = Utils.BenchmarkTimer()
      def wrapped(*args):
         self.benchmarks[func].startRecord()
         ret = func(*args)
         self.benchmarks[func].stopRecord()
         return ret
      return wrapped

   def bench(self, tick):
      if tick % 100 == 0:
         for k, benchmark in self.benchmarks.items():
            bench = benchmark.benchmark()
            print(k.__func__.__name__, 'Tick: ', tick,
                  ', Benchmark: ', bench, ', FPS: ', 1/bench)
 




from pdb import set_trace as T
import numpy as np

from nmmo.lib import overlay
from nmmo.lib.colors import Neon
from nmmo.systems import combat


class OverlayRegistry:
   def __init__(self, config, realm):
      '''Manager class for overlays

      Args:
          config: A Config object
          realm: An environment
      '''
      self.initialized = False

      self.config = config
      self.realm  = realm

      self.overlays = {
              'counts':     Counts,
              'skills':     Skills,
              'wilderness': Wilderness}


   def init(self, *args):
      self.initialized = True
      for cmd, overlay in self.overlays.items():
         self.overlays[cmd] = overlay(self.config, self.realm, *args)
      return self

   def step(self, obs, pos, cmd):
      '''Per-tick overlay updates

      Args:
          obs: Observation returned by the environment
          pos: Client camera focus position
          cmd: User command returned by the client
      '''
      if not self.initialized:
          self.init()

      self.realm.overlayPos = pos
      for overlay in self.overlays.values():
          overlay.update(obs)

      if cmd in self.overlays:
          self.overlays[cmd].register(obs)

class Overlay:
   '''Define a overlay for visualization in the client

   Overlays are color images of the same size as the game map.
   They are rendered over the environment with transparency and
   can be used to gain insight about agent behaviors.'''
   def __init__(self, config, realm, *args):
      '''
      Args:
          config: A Config object
          realm: An environment
      '''
      self.config     = config
      self.realm      = realm

      self.size       = config.MAP_SIZE
      self.values     = np.zeros((self.size, self.size))

   def update(self, obs):
       '''Compute per-tick updates to this overlay. Override per overlay.

       Args:
           obs: Observation returned by the environment
       '''
       pass

   def register(self):
       '''Compute the overlay and register it within realm. Override per overlay.'''
       pass

class Skills(Overlay):
   def __init__(self, config, realm, *args):
      '''Indicates whether agents specialize in foraging or combat'''
      super().__init__(config, realm)
      self.nSkills = 2

      self.values  = np.zeros((self.size, self.size, self.nSkills))

   def update(self, obs):
      '''Computes a count-based exploration map by painting
      tiles as agents walk over them'''
      for entID, agent in self.realm.realm.players.items():
         r, c = agent.base.pos

         skillLvl  = (agent.skills.food.level.val + agent.skills.water.level.val)/2.0
         combatLvl = combat.level(agent.skills)

         if skillLvl == 10 and combatLvl == 3:
            continue

         self.values[r, c, 0] = skillLvl
         self.values[r, c, 1] = combatLvl

   def register(self, obs):
      values = np.zeros((self.size, self.size, self.nSkills))
      for idx in range(self.nSkills):
         ary  = self.values[:, :, idx]
         vals = ary[ary != 0]
         mean = np.mean(vals)
         std  = np.std(vals)
         if std == 0:
            std = 1

         values[:, :, idx] = (ary - mean) / std
         values[ary == 0] = 0

      colors    = np.array([Neon.BLUE.rgb, Neon.BLOOD.rgb])
      colorized = np.zeros((self.size, self.size, 3))
      amax      = np.argmax(values, -1)

      for idx in range(self.nSkills):
         colorized[amax == idx] = colors[idx] / 255
         colorized[values[:, :, idx] == 0] = 0

      self.realm.register(colorized)

class Counts(Overlay):
   def __init__(self, config, realm, *args):
      super().__init__(config, realm)
      self.values = np.zeros((self.size, self.size, config.PLAYER_POLICIES))

   def update(self, obs):
      '''Computes a count-based exploration map by painting
      tiles as agents walk over them'''
      for entID, agent in self.realm.realm.players.items():
         pop  = agent.base.population.val
         r, c = agent.base.pos
         self.values[r, c][pop] += 1

   def register(self, obs):
      colors    = self.realm.realm.players.palette.colors
      colors    = np.array([colors[pop].rgb
            for pop in range(self.config.PLAYER_POLICIES)])

      colorized = self.values[:, :, :, None] * colors / 255
      colorized = np.sum(colorized, -2)
      countSum  = np.sum(self.values[:, :], -1)
      data      = overlay.norm(countSum)[..., None]

      countSum[countSum==0] = 1
      colorized = colorized * data / countSum[..., None]

      self.realm.register(colorized)

class Wilderness(Overlay):
   def init(self):
      '''Computes the local wilderness level'''
      data = np.zeros((self.size, self.size))
      for r in range(self.size):
         for c in range(self.size):
            data[r, c] = combat.wilderness(self.config, (r, c))

      self.wildy = overlay.twoTone(data, preprocess='clip', invert=True, periods=5)

   def register(self, obs):
      if not hasattr(self, 'wildy'):
         print('Initializing Wilderness')
         self.init()

      self.realm.register(self.wildy)



from pdb import set_trace as T
import numpy as np

import nmmo
from nmmo.systems import skill, droptable, combat, equipment, inventory
from nmmo.lib import material, utils

class Resources:
   def __init__(self, ent):
      self.health = nmmo.Serialized.Entity.Health(ent.dataframe, ent.entID)
      self.water  = nmmo.Serialized.Entity.Water( ent.dataframe, ent.entID)
      self.food   = nmmo.Serialized.Entity.Food(  ent.dataframe, ent.entID)

   def update(self, realm, entity, actions):
      config = realm.config

      if not config.RESOURCE_SYSTEM_ENABLED:
         return

      self.water.max = config.RESOURCE_BASE
      self.food.max  = config.RESOURCE_BASE

      regen  = config.RESOURCE_HEALTH_RESTORE_FRACTION
      thresh = config.RESOURCE_HEALTH_REGEN_THRESHOLD

      food_thresh  = self.food  > thresh * config.RESOURCE_BASE
      water_thresh = self.water > thresh * config.RESOURCE_BASE

      if food_thresh and water_thresh:
          restore = np.floor(self.health.max * regen)
          self.health.increment(restore)

      if self.food.empty:
          self.health.decrement(config.RESOURCE_STARVATION_RATE)

      if self.water.empty:
          self.health.decrement(config.RESOURCE_DEHYDRATION_RATE)

   def packet(self):
      data = {}
      data['health'] = self.health.packet()
      data['food']   = self.food.packet()
      data['water']  = self.water.packet()
      return data

class Status:
   def __init__(self, ent):
      self.config = ent.config
      self.freeze = nmmo.Serialized.Entity.Freeze(ent.dataframe, ent.entID)

   def update(self, realm, entity, actions):
      self.freeze.decrement()

   def packet(self):
      data = {}
      data['freeze'] = self.freeze.val
      return data

class History:
   def __init__(self, ent):
      self.actions = {}
      self.attack  = None
  
      self.origPos     = ent.pos
      self.exploration = 0
      self.playerKills = 0

      self.damage_received = 0
      self.damage_inflicted = 0

      self.damage    = nmmo.Serialized.Entity.Damage(   ent.dataframe, ent.entID)
      self.timeAlive = nmmo.Serialized.Entity.TimeAlive(ent.dataframe, ent.entID)

      self.lastPos = None

   def update(self, realm, entity, actions):
      self.attack  = None
      self.damage.update(0)

      self.actions = {}
      if entity.entID in actions:
          self.actions = actions[entity.entID]
 
      exploration      = utils.linf(entity.pos, self.origPos)
      self.exploration = max(exploration, self.exploration)

      self.timeAlive.increment()

   def packet(self):
      data = {}
      data['damage']    = self.damage.val
      data['timeAlive'] = self.timeAlive.val
      data['damage_inflicted'] = self.damage_inflicted
      data['damage_received'] = self.damage_received

      if self.attack is not None:
         data['attack'] = self.attack

      actions = {}
      for atn, args in self.actions.items():
         atn_packet = {}

         #Avoid recursive player packet
         if atn.__name__ == 'Attack':
             continue

         for key, val in args.items():
            if hasattr(val, 'packet'):
               atn_packet[key.__name__] = val.packet
            else:
               atn_packet[key.__name__] = val.__name__
         actions[atn.__name__] = atn_packet
      data['actions'] = actions

      return data

class Base:
   def __init__(self, ent, pos, iden, name, color, pop):
      self.name  = name + str(iden)
      self.color = color
      r, c       = pos

      self.r          = nmmo.Serialized.Entity.R(ent.dataframe, ent.entID, r)
      self.c          = nmmo.Serialized.Entity.C(ent.dataframe, ent.entID, c)

      self.population = nmmo.Serialized.Entity.Population(ent.dataframe, ent.entID, pop)
      self.self       = nmmo.Serialized.Entity.Self(      ent.dataframe, ent.entID, 1)
      self.identity   = nmmo.Serialized.Entity.ID(        ent.dataframe, ent.entID, ent.entID)
      self.level      = nmmo.Serialized.Entity.Level(     ent.dataframe, ent.entID, 3)
      self.item_level = nmmo.Serialized.Entity.ItemLevel( ent.dataframe, ent.entID, 0)
      self.gold       = nmmo.Serialized.Entity.Gold(      ent.dataframe, ent.entID, 0)
      self.comm       = nmmo.Serialized.Entity.Comm(      ent.dataframe, ent.entID, 0)

      ent.dataframe.init(nmmo.Serialized.Entity, ent.entID, (r, c))

   def update(self, realm, entity, actions):
      self.level.update(combat.level(entity.skills))

      if realm.config.EQUIPMENT_SYSTEM_ENABLED:
         self.item_level.update(entity.equipment.total(lambda e: e.level))

      if realm.config.EXCHANGE_SYSTEM_ENABLED:
         self.gold.update(entity.inventory.gold.quantity.val)

   @property
   def pos(self):
      return self.r.val, self.c.val

   def packet(self):
      data = {}

      data['r']          = self.r.val
      data['c']          = self.c.val
      data['name']       = self.name
      data['level']      = self.level.val
      data['item_level'] = self.item_level.val
      data['color']      = self.color.packet()
      data['population'] = self.population.val
      data['self']       = self.self.val

      return data

class Entity:
   def __init__(self, realm, pos, iden, name, color, pop):
      self.realm        = realm
      self.dataframe    = realm.dataframe
      self.config       = realm.config

      self.policy       = name
      self.entID        = iden
      self.repr         = None
      self.vision       = 5

      self.attacker     = None
      self.target       = None
      self.closest      = None
      self.spawnPos     = pos

      self.attackerID = nmmo.Serialized.Entity.AttackerID(self.dataframe, self.entID, 0)

      #Submodules
      self.base      = Base(self, pos, iden, name, color, pop)
      self.status    = Status(self)
      self.history   = History(self)
      self.resources = Resources(self)

      self.inventory = inventory.Inventory(realm, self)

   def packet(self):
      data = {}

      data['status']    = self.status.packet()
      data['history']   = self.history.packet()
      data['inventory'] = self.inventory.packet()
      data['alive']     = self.alive

      return data

   def update(self, realm, actions):
      '''Update occurs after actions, e.g. does not include history'''
      if self.history.damage == 0:
         self.attacker = None
         self.attackerID.update(0)

      self.base.update(realm, self, actions)
      self.status.update(realm, self, actions)
      self.history.update(realm, self, actions)

   def receiveDamage(self, source, dmg):
      self.history.damage_received += dmg
      self.history.damage.update(dmg)
      self.resources.health.decrement(dmg)

      if self.alive:
          return True

      if source is None:
          return True 

      if not source.isPlayer:
          return True 

      return False

   def applyDamage(self, dmg, style):
      self.history.damage_inflicted += dmg

   @property
   def pos(self):
      return self.base.pos

   @property
   def alive(self):
      if self.resources.health.empty:
         return False

      return True

   @property
   def isPlayer(self) -> bool:
      return False

   @property
   def isNPC(self) -> bool:
      return False

   @property
   def level(self) -> int:
       melee  = self.skills.melee.level.val
       ranged = self.skills.range.level.val
       mage   = self.skills.mage.level.val

       return int(max(melee, ranged, mage))



from pdb import set_trace as T

from collections import defaultdict, deque
from queue import PriorityQueue

import inspect
import logging
import inspect

import math

class Offer:
   def __init__(self, seller, item):
      self.seller = seller
      self.item   = item

   '''
   def __lt__(self, offer):
      return self.price < offer.price
      
   def __le__(self, offer):
      return self.price <= offer.price

   def __eq__(self, offer):
      return self.price == offer.price

   def __ne__(self, offer):
      return self.price != offer.price

   def __gt__(self, offer):
      return self.price > offer.price

   def __ge__(self, offer):
      return self.price >= offer.price
   '''

#Why is the api so weird...
class Queue(deque):
   def __init__(self):
      super().__init__()
      self.price = None

   def push(self, x):
      self.appendleft(x)
   
   def peek(self):
      if len(self) > 0:
         return self[-1]
      return None

class ItemListings:
   def __init__(self):
      self.listings    = PriorityQueue()
      self.placeholder = None
      self.item_number = 0
      self.alpha       = 0.01
      self.volume = 0

      self.step()

   def step(self):
      #self.volume = 0
      pass

   @property
   def price(self):
      if not self.supply:
          return

      price, item_number, seller = self.listings.get()
      self.listings.put((price, item_number, seller))
      return price

   @property
   def supply(self):
      return self.listings.qsize()

   @property
   def empty(self):
      return self.listings.empty()

   def buy(self, buyer, max_price):
      if not self.supply:
         return

      price, item_number, seller = self.listings.get()

      if price > max_price or price > buyer.inventory.gold.quantity.val:
         self.listings.put((price, item_number, seller))
         return

      seller.inventory.gold.quantity += price
      buyer.inventory.gold.quantity  -= price

      buyer.buys   += 1
      seller.sells += 1
      self.volume  += 1
      return price
         
   def sell(self, seller, price):
      if price == 1 and not self.empty:
         seller.inventory.gold.quantity += 1
      else:
         self.listings.put((price, self.item_number, seller))
         self.item_number += 1

      #print('Sell {}: {}'.format(item.__class__.__name__, price))

class Exchange:
   def __init__(self):
      self.item_listings = defaultdict(ItemListings)

   @property
   def dataframeKeys(self):
      keys = []
      for listings in self.item_listings.values():
         if listings.placeholder:
            keys.append(listings.placeholder.instanceID)

      return keys

   @property
   def dataframeVals(self):
      vals = []
      for listings in self.item_listings.values():
         if listings.placeholder:
            vals.append(listings.placeholder)

      return vals

   @property
   def packet(self):
      packet = {}
      for (item_cls, level), listings in self.item_listings.items():
          key = f'{item_cls.__name__}_{level}'

          item = listings.placeholder
          if item is None:
              continue

          packet[key] = {
                  'price': listings.price,
                  'supply': listings.supply}

      return packet

   def step(self):
      for item, listings in self.item_listings.items():
         listings.step()

   def available(self, item):
      return self.item_listings[item].available()

   def buy(self, realm, buyer, item):
      assert isinstance(item, object), f'{item} purchase is not an Item instance'
      assert item.quantity.val == 1, f'{item} purchase has quantity {item.quantity.val}'

      #TODO: Handle ammo stacks
      if not buyer.inventory.space:
         return

      config       = realm.config
      level        = item.level.val

      #Agents may try to buy an item at the same time
      #Therefore the price has to be semi-variable
      price        = item.price.val
      max_price    = 1.1 * price

      item         = type(item)
      listings_key = (item, level)
      listings     = self.item_listings[listings_key]

      price = listings.buy(buyer, max_price)
      if price is not None:
         buyer.inventory.receive(listings.placeholder)

         if ((config.LOG_MILESTONES and realm.quill.milestone.log_max(f'Buy_{item.__name__}', level)) or 
               (config.LOG_EVENTS and realm.quill.event.log(f'Buy_{item.__name__}', level))) and config.LOG_VERBOSE:
            logging.info(f'EXCHANGE: Bought level {level} {item.__name__} for {price} gold')
         if ((config.LOG_MILESTONES and realm.quill.milestone.log_max(f'Transaction_Amount', price)) or 
               (config.LOG_EVENTS and realm.quill.event.log(f'Transaction_Amount', price))) and config.LOG_VERBOSE:
            logging.info(f'EXCHANGE: Transaction of {price} gold (level {level} {item.__name__})')
 
         #Update placeholder
         listings.placeholder = None
         if listings.supply:
            listings.placeholder = item(realm, level, price=listings.price)
            
   def sell(self, realm, seller, item, price):
      assert isinstance(item, object), f'{item} for sale is not an Item instance'
      assert item in seller.inventory, f'{item} for sale is not in {seller} inventory'
      assert item.quantity.val > 0, f'{item} for sale has quantity {item.quantity.val}'

      if not item.tradable.val:
         return

      config   = realm.config
      level    = item.level.val

      #Remove from seller
      seller.inventory.remove(item, quantity=1)
      item = type(item)

     
      if ((config.LOG_MILESTONES and realm.quill.milestone.log_max(f'Sell_{item.__name__}', level)) or (config.LOG_EVENTS and realm.quill.event.log(f'Sell_{item.__name__}', level))) and config.LOG_VERBOSE:
         logging.info(f'EXCHANGE: Offered level {level} {item.__name__} for {price} gold')

      listings_key  = (item, level)
      listings      = self.item_listings[listings_key]
      current_price = listings.price

      #Update obs placeholder item
      if listings.placeholder is None or (current_price is not None and price < current_price):
         listings.placeholder = item(realm, level, price=price)

      #print('{} Sold {} x {} for {} ea.'.format(seller.base.name, quantity, item.__name__, price))
      listings.sell(seller, price)



from pdb import set_trace as T
import numpy as np

import random

import nmmo
from nmmo.entity import entity
from nmmo.systems import combat, equipment, ai, combat, skill
from nmmo.lib.colors import Neon
from nmmo.systems import item as Item
from nmmo.systems import droptable
from nmmo.io import action as Action


class Equipment:
   def __init__(self, total,
           melee_attack, range_attack, mage_attack,
           melee_defense, range_defense, mage_defense):

      self.level         = total
      self.ammunition    = None

      self.melee_attack  = melee_attack
      self.range_attack  = range_attack
      self.mage_attack   = mage_attack
      self.melee_defense = melee_defense
      self.range_defense = range_defense
      self.mage_defense  = mage_defense

   def total(self, getter):
      return getter(self)

   @property
   def packet(self):
      packet = {}

      packet['item_level']    = self.total

      packet['melee_attack']  = self.melee_attack
      packet['range_attack']  = self.range_attack
      packet['mage_attack']   = self.mage_attack
      packet['melee_defense'] = self.melee_defense
      packet['range_defense'] = self.range_defense
      packet['mage_defense']  = self.mage_defense

      return packet


class NPC(entity.Entity):
   def __init__(self, realm, pos, iden, name, color, pop):
      super().__init__(realm, pos, iden, name, color, pop)
      self.skills = skill.Combat(realm, self)
      self.realm = realm
      

   def update(self, realm, actions):
      super().update(realm, actions)

      if not self.alive:
         return

      self.resources.health.increment(1)
      self.lastAction = actions

   def receiveDamage(self, source, dmg):
       if super().receiveDamage(source, dmg):
           return True

       for item in self.droptable.roll(self.realm, self.level):
           if source.inventory.space:
               source.inventory.receive(item)

   @staticmethod
   def spawn(realm, pos, iden):
      config = realm.config

      # Select AI Policy
      danger = combat.danger(config, pos)
      if danger >= config.NPC_SPAWN_AGGRESSIVE:
         ent = Aggressive(realm, pos, iden)
      elif danger >= config.NPC_SPAWN_NEUTRAL:
         ent = PassiveAggressive(realm, pos, iden)
      elif danger >= config.NPC_SPAWN_PASSIVE:
         ent = Passive(realm, pos, iden)
      else:
         return

      ent.spawn_danger = danger

      # Select combat focus
      style = random.choice((Action.Melee, Action.Range, Action.Mage))
      ent.skills.style = style

      # Compute level
      level = 0
      if config.PROGRESSION_SYSTEM_ENABLED:
          level_min = config.NPC_LEVEL_MIN
          level_max = config.NPC_LEVEL_MAX
          level     = int(danger * (level_max - level_min) + level_min)

          # Set skill levels
          if style == Action.Melee:
              ent.skills.melee.setExpByLevel(level)
          elif style == Action.Range:
              ent.skills.range.setExpByLevel(level)
          elif style == Action.Mage:
              ent.skills.mage.setExpByLevel(level)

      # Gold
      if config.EXCHANGE_SYSTEM_ENABLED:
          ent.inventory.gold.quantity.update(level)

      ent.droptable = droptable.Standard()

      # Equipment to instantiate
      if config.EQUIPMENT_SYSTEM_ENABLED:
          lvl     = level - random.random()
          ilvl    = int(5 * lvl)

          offense = int(config.NPC_BASE_DAMAGE + lvl*config.NPC_LEVEL_DAMAGE)
          defense = int(config.NPC_BASE_DEFENSE + lvl*config.NPC_LEVEL_DEFENSE)

          ent.equipment = Equipment(ilvl, offense, offense, offense, defense, defense, defense)

          armor =  [Item.Hat, Item.Top, Item.Bottom]
          ent.droptable.add(random.choice(armor))

      if config.PROFESSION_SYSTEM_ENABLED:
         tools =  [Item.Rod, Item.Gloves, Item.Pickaxe, Item.Chisel, Item.Arcane]
         ent.droptable.add(random.choice(tools))

      return ent 

   def packet(self):
      data = super().packet()

      data['base']     = self.base.packet()      
      data['skills']   = self.skills.packet()      
      data['resource'] = {'health': self.resources.health.packet()}

      return data

   @property
   def isNPC(self) -> bool:
      return True

class Passive(NPC):
   def __init__(self, realm, pos, iden):
      super().__init__(realm, pos, iden, 'Passive', Neon.GREEN, -1)
      self.dataframe.init(nmmo.Serialized.Entity, iden, pos)

   def decide(self, realm):
      return ai.policy.passive(realm, self)

class PassiveAggressive(NPC):
   def __init__(self, realm, pos, iden):
      super().__init__(realm, pos, iden, 'Neutral', Neon.ORANGE, -2)
      self.dataframe.init(nmmo.Serialized.Entity, iden, pos)

   def decide(self, realm):
      return ai.policy.neutral(realm, self)

class Aggressive(NPC):
   def __init__(self, realm, pos, iden):
      super().__init__(realm, pos, iden, 'Hostile', Neon.RED, -3)
      self.dataframe.init(nmmo.Serialized.Entity, iden, pos)

   def decide(self, realm):
      return ai.policy.hostile(realm, self)



from pdb import set_trace as T
import numpy as np

import nmmo
from nmmo.lib import material

class Tile:
   def __init__(self, config, realm, r, c):
      self.config = config
      self.realm  = realm

      self.serialized = 'R{}-C{}'.format(r, c)

      self.r     = nmmo.Serialized.Tile.R(realm.dataframe, self.serial, r)
      self.c     = nmmo.Serialized.Tile.C(realm.dataframe, self.serial, c)
      self.nEnts = nmmo.Serialized.Tile.NEnts(realm.dataframe, self.serial)
      self.index = nmmo.Serialized.Tile.Index(realm.dataframe, self.serial, 0)

      realm.dataframe.init(nmmo.Serialized.Tile, self.serial, (r, c))

   @property
   def serial(self):
      return self.serialized

   @property
   def repr(self):
      return ((self.r, self.c))

   @property
   def pos(self):
      return self.r.val, self.c.val

   @property
   def habitable(self):
      return self.mat in material.Habitable

   @property
   def vacant(self):
      return len(self.ents) == 0 and self.habitable

   @property
   def occupied(self):
      return not self.vacant

   @property
   def impassible(self):
      return self.mat in material.Impassible

   @property
   def lava(self):
      return self.mat == material.Lava

   def reset(self, mat, config):
      self.state  = mat(config)
      self.mat    = mat(config)

      self.depleted = False
      self.tex      = mat.tex
      self.ents     = {}

      self.nEnts.update(0)
      self.index.update(self.state.index)
 
   def addEnt(self, ent):
      assert ent.entID not in self.ents
      self.nEnts.update(1)
      self.ents[ent.entID] = ent

   def delEnt(self, entID):
      assert entID in self.ents
      self.nEnts.update(0)
      del self.ents[entID]

   def step(self):
      if not self.depleted or np.random.rand() > self.mat.respawn:
         return

      self.depleted = False
      self.state = self.mat

      self.index.update(self.state.index)

   def harvest(self, deplete):
      if __debug__:
          assert not self.depleted, f'{self.state} is depleted'
          assert self.state in material.Harvestable, f'{self.state} not harvestable'

      if deplete:
          self.depleted = True
          self.state    = self.mat.deplete(self.config)
          self.index.update(self.state.index)

      return self.mat.harvest()



from pdb import set_trace as T
import numpy as np

import nmmo
from nmmo.systems.ai import move, attack, utils

def update(entity):
   '''Update validity of tracked entities'''
   if not utils.validTarget(entity, entity.attacker, entity.vision):
      entity.attacker = None
   if not utils.validTarget(entity, entity.target, entity.vision):
      entity.target = None
   if not utils.validTarget(entity, entity.closest, entity.vision):
      entity.closest = None

   if entity.__class__.__name__ != 'Player':
      return

   if not utils.validResource(entity, entity.food, entity.vision):
      entity.food = None
   if not utils.validResource(entity, entity.water, entity.vision):
      entity.water = None
      
def pathfind(realm, actions, entity, target):
   actions[nmmo.action.Move]   = {nmmo.action.Direction: move.pathfind(realm.map.tiles, entity, target)}

def explore(realm, actions, entity):
   sz   = realm.config.TERRAIN_SIZE
   r, c = entity.pos

   spawnR, spawnC = entity.spawnPos
   centR, centC   = sz//2, sz//2

   vR, vC = centR-spawnR, centC-spawnC

   mmag = max(abs(vR), abs(vC))
   rr   = r + int(np.round(entity.vision*vR/mmag))
   cc   = c + int(np.round(entity.vision*vC/mmag))

   tile = realm.map.tiles[rr, cc]
   pathfind(realm, actions, entity, tile)

def explore(config, ob, actions, spawnR, spawnC):
   vision = config.NSTIM
   sz     = config.TERRAIN_SIZE
   Entity = nmmo.Serialized.Entity
   Tile   = nmmo.Serialized.Tile

   agent  = ob.agent
   r      = utils.Observation.attribute(agent, Entity.R)
   c      = utils.Observation.attribute(agent, Entity.C)

   centR, centC   = sz//2, sz//2

   vR, vC = centR-spawnR, centC-spawnC

   mmag = max(abs(vR), abs(vC))
   rr   = int(np.round(vision*vR/mmag))
   cc   = int(np.round(vision*vC/mmag))

   pathfind(config, ob, actions, rr, cc)

def meander(realm, actions, entity):
   actions[nmmo.action.Move] = {nmmo.action.Direction: move.habitable(realm.map.tiles, entity)}

def evade(realm, actions, entity):
   actions[nmmo.action.Move] = {nmmo.action.Direction: move.antipathfind(realm.map.tiles, entity, entity.attacker)}

def hunt(realm, actions, entity):
   #Move args
   distance = utils.distance(entity, entity.target)

   direction = None
   if distance == 0:
      direction = move.random()
   elif distance > 1:
      direction = move.pathfind(realm.map.tiles, entity, entity.target)

   if direction is not None:
      actions[nmmo.action.Move] = {nmmo.action.Direction: direction}

   attack(realm, actions, entity)

def attack(realm, actions, entity):
   distance = utils.lInfty(entity.pos, entity.target.pos)
   if distance > entity.skills.style.attackRange(realm.config):
      return

   actions[nmmo.action.Attack] = {
         nmmo.action.Style: entity.skills.style,
         nmmo.action.Target: entity.target}




from pdb import set_trace as T
import numpy as np


class SequentialLoader:
    '''config.PLAYER_LOADER that spreads out agent populations'''
    def __init__(self, config):
        items = config.PLAYERS
        for idx, itm in enumerate(items):
           itm.policyID = idx 

        self.items = items
        self.idx   = -1

    def __iter__(self):
        return self

    def __next__(self):
        self.idx = (self.idx + 1) % len(self.items)
        return self.idx, self.items[self.idx]

class TeamLoader:
    '''config.PLAYER_LOADER that loads agent populations adjacent'''
    def __init__(self, config):
        items = config.PLAYERS
        self.team_size = config.PLAYER_N // len(items)

        for idx, itm in enumerate(items):
           itm.policyID = idx 

        self.items = items
        self.idx   = -1

    def __iter__(self):
        return self

    def __next__(self):
        self.idx += 1
        team_idx  = self.idx // self.team_size
        return team_idx, self.items[team_idx]


def spawn_continuous(config):
    '''Generates spawn positions for new agents

    Randomly selects spawn positions around
    the borders of the square game map

    Returns:
        tuple(int, int):

    position:
        The position (row, col) to spawn the given agent
    '''
    #Spawn at edges
    mmax = config.MAP_CENTER + config.MAP_BORDER
    mmin = config.MAP_BORDER

    var  = np.random.randint(mmin, mmax)
    fixed = np.random.choice([mmin, mmax])
    r, c = int(var), int(fixed)
    if np.random.rand() > 0.5:
        r, c = c, r 
    return (r, c)

def old_spawn_concurrent(config):
    '''Generates spawn positions for new agents

    Evenly spaces agents around the borders
    of the square game map

    Returns:
        tuple(int, int):

    position:
        The position (row, col) to spawn the given agent
    '''
 
    left   = config.MAP_BORDER
    right  = config.MAP_CENTER + config.MAP_BORDER
    rrange = np.arange(left+2, right, 4).tolist()

    assert not config.MAP_CENTER % 4
    per_side = config.MAP_CENTER // 4
  
    lows   = (left+np.zeros(per_side, dtype=np.int)).tolist()
    highs  = (right+np.zeros(per_side, dtype=np.int)).tolist()

    s1     = list(zip(rrange, lows))
    s2     = list(zip(lows, rrange))
    s3     = list(zip(rrange, highs))
    s4     = list(zip(highs, rrange))

    ret = s1 + s2 + s3 + s4

    # Shuffle needs porting to competition version
    np.random.shuffle(ret)

    return ret

def spawn_concurrent(config):
    '''Generates spawn positions for new agents

    Evenly spaces agents around the borders
    of the square game map

    Returns:
        tuple(int, int):

    position:
        The position (row, col) to spawn the given agent
    '''
    team_size = config.PLAYER_TEAM_SIZE
    team_n = len(config.PLAYERS)
    teammate_sep = config.PLAYER_SPAWN_TEAMMATE_DISTANCE

    # Number of total border tiles
    total_tiles = 4 * config.MAP_CENTER 

    # Number of tiles, including within-team sep, occupied by each team
    tiles_per_team = teammate_sep*(team_size-1) + team_size

    # Number of total tiles dedicated to separating teams
    buffer_tiles = 0
    if team_n > 1:
        buffer_tiles = total_tiles - tiles_per_team*team_n

    # Number of tiles between teams
    team_sep = buffer_tiles // team_n

    # Accounts for lava borders in coord calcs
    left = config.MAP_BORDER
    right = config.MAP_CENTER + config.MAP_BORDER
    lows = config.MAP_CENTER * [left]
    highs = config.MAP_CENTER * [right]
    inc = list(range(config.MAP_BORDER, config.MAP_CENTER+config.MAP_BORDER))

    # All edge tiles in order
    sides = []
    sides += list(zip(lows, inc))
    sides += list(zip(inc, highs))
    sides += list(zip(highs, inc[::-1]))
    sides += list(zip(inc[::-1], lows))

    # Space across and within teams
    spawn_positions = []
    for idx in range(team_sep//2, len(sides), tiles_per_team+team_sep):
        for offset in list(range(0,  tiles_per_team, teammate_sep+1)):
            if len(spawn_positions) >= config.PLAYER_N:
                continue

            pos = sides[idx + offset]
            spawn_positions.append(pos)

    return spawn_positions




from pdb import set_trace as T

from nmmo.systems import item, droptable

class Material:
   capacity    = 0
   tool        = None
   table       = None

   def __init__(self, config):
      pass

   def __eq__(self, mtl):
      return self.index == mtl.index

   def __equals__(self, mtl):
      return self == mtl

   def harvest(self):
      return self.__class__.table

class Lava(Material):
   tex   = 'lava'
   index = 0

class Water(Material):
   tex   = 'water'
   index = 1

   table = droptable.Empty()

   def __init__(self, config):
       self.deplete = __class__
       self.respawn  = 1.0

class Grass(Material):
   tex   = 'grass'
   index = 2

class Scrub(Material):
   tex   = 'scrub'
   index = 3

class Forest(Material):
   tex   = 'forest'
   index = 4

   deplete = Scrub
   table = droptable.Empty()

   def __init__(self, config):
      if config.RESOURCE_SYSTEM_ENABLED:
         self.capacity = config.RESOURCE_FOREST_CAPACITY
         self.respawn  = config.RESOURCE_FOREST_RESPAWN

class Stone(Material):
   tex   = 'stone'
   index = 5

class Slag(Material):
   tex   = 'slag'
   index = 6

class Ore(Material):
   tex   = 'ore'
   index = 7

   deplete = Slag
   tool    = item.Pickaxe

   def __init__(self, config):
       cls = self.__class__
       if cls.table is None:
           cls.table = droptable.Standard()
           cls.table.add(item.Scrap)

           if config.EQUIPMENT_SYSTEM_ENABLED:
               cls.table.add(item.Wand, prob=config.WEAPON_DROP_PROB)

       self.capacity = config.PROFESSION_ORE_CAPACITY
       self.respawn  = config.PROFESSION_ORE_RESPAWN

   tool    = item.Pickaxe
   deplete = Slag

class Stump(Material):
   tex   = 'stump'
   index = 8

class Tree(Material):
   tex   = 'tree'
   index = 9

   deplete = Stump
   tool    = item.Chisel

   def __init__(self, config):
      cls = self.__class__
      if cls.table is None:
           cls.table = droptable.Standard()
           cls.table.add(item.Shaving)
           if config.EQUIPMENT_SYSTEM_ENABLED:
               cls.table.add(item.Sword, prob=config.WEAPON_DROP_PROB)

      self.capacity = config.PROFESSION_TREE_CAPACITY
      self.respawn  = config.PROFESSION_TREE_RESPAWN

class Fragment(Material):
   tex   = 'fragment'
   index = 10

class Crystal(Material):
   tex   = 'crystal'
   index = 11

   deplete = Fragment
   tool    = item.Arcane

   def __init__(self, config):
      cls = self.__class__
      if cls.table is None:
          cls.table = droptable.Standard()
          cls.table.add(item.Shard)
          if config.EQUIPMENT_SYSTEM_ENABLED:
              cls.table.add(item.Bow, prob=config.WEAPON_DROP_PROB)

      if config.RESOURCE_SYSTEM_ENABLED:
          self.capacity = config.PROFESSION_CRYSTAL_CAPACITY
          self.respawn  = config.PROFESSION_CRYSTAL_RESPAWN

class Weeds(Material):
   tex   = 'weeds'
   index = 12

class Herb(Material):
   tex   = 'herb'
   index = 13

   deplete = Weeds
   tool    = item.Gloves

   table   = droptable.Standard()
   table.add(item.Poultice)

   def __init__(self, config):
      if config.RESOURCE_SYSTEM_ENABLED:
         self.capacity = config.PROFESSION_HERB_CAPACITY
         self.respawn  = config.PROFESSION_HERB_RESPAWN

class Ocean(Material):
   tex   = 'ocean'
   index = 14

class Fish(Material):
   tex   = 'fish'
   index = 15

   deplete = Ocean
   tool    = item.Rod

   table   = droptable.Standard()
   table.add(item.Ration)

   def __init__(self, config):
      if config.RESOURCE_SYSTEM_ENABLED:
         self.capacity = config.PROFESSION_FISH_CAPACITY
         self.respawn  = config.PROFESSION_FISH_RESPAWN

class Meta(type):
   def __init__(self, name, bases, dict):
      self.indices = {mtl.index for mtl in self.materials}

   def __iter__(self):
      yield from self.materials

   def __contains__(self, mtl):
      if isinstance(mtl, Material):
         mtl = type(mtl)
      if isinstance(mtl, type):
         return mtl in self.materials
      return mtl in self.indices

class All(metaclass=Meta):
   '''List of all materials'''
   materials = {
      Lava, Water, Grass, Scrub, Forest,
      Stone, Slag, Ore, Stump, Tree,
      Fragment, Crystal, Weeds, Herb, Ocean, Fish}

class Impassible(metaclass=Meta):
   '''Materials that agents cannot walk through'''
   materials = {Lava, Water, Stone, Ocean, Fish}

class Habitable(metaclass=Meta):
   '''Materials that agents cannot walk on'''
   materials = {Grass, Scrub, Forest, Ore, Slag, Tree, Stump, Crystal, Fragment, Herb, Weeds}

class Harvestable(metaclass=Meta):
   '''Materials that agents can harvest'''
   materials = {Water, Forest, Ore, Tree, Crystal, Herb, Fish}



import numpy as np

class Empty():
    def roll(self, realm, level):
        return []

class Fixed():
    def __init__(self, item, amount=1):
        self.item = item
        self.amount = amount

    def roll(self, realm, level):
        return [self.item(realm, level, amount=amount)]

class Drop:
   def __init__(self, item, amount, prob):
      self.item = item
      self.amount = amount
      self.prob = prob

   def roll(self, realm, level):
      if np.random.rand() < self.prob:
         return self.item(realm, level, quantity=self.amount)

class Standard:
   def __init__(self):
      self.drops = []

   def add(self, item, quant=1, prob=1.0):
      self.drops += [Drop(item, quant, prob)]

   def roll(self, realm, level):
      ret = []
      for e in self.drops:
         drop = e.roll(realm, level)
         if drop is not None:
            ret += [drop]
      return ret

class Empty(Standard):
    def roll(self, realm, level):
        return []

class Ammunition(Standard):
    def __init__(self, item):
        self.item = item

    def roll(self, realm, level):
        return [self.item(realm, level)]

class Consumable(Standard):
    def __init__(self, item):
        self.item = item

    def roll(self, realm, level):
        return [self.item(realm, level)]



#Various utilities for managing combat, including hit/damage

from pdb import set_trace as T

import numpy as np
import logging

from nmmo.systems import skill as Skill
from nmmo.systems import item as Item

def level(skills):
    return max(e.level.val for e in skills.skills)

def damage_multiplier(config, skill, targ):
    skills = [targ.skills.melee, targ.skills.range, targ.skills.mage]
    exp    = [s.exp for s in skills]

    if max(exp) == min(exp):
        return 1.0

    idx    = np.argmax([exp])
    targ   = skills[idx]

    if type(skill) == targ.weakness:
        return config.COMBAT_WEAKNESS_MULTIPLIER

    return 1.0

def attack(realm, player, target, skillFn):
    config       = player.config
    skill        = skillFn(player)
    skill_type   = type(skill)
    skill_name   = skill_type.__name__

    # Attacker and target levels
    player_level = skill.level.val
    target_level = level(target.skills)

    # Ammunition usage
    ammunition = player.equipment.ammunition
    if ammunition is not None:
        ammunition.fire(player)

    # Per-style offense/defense
    level_damage = 0
    if skill_type == Skill.Melee:
        base_damage  = config.COMBAT_MELEE_DAMAGE

        if config.PROGRESSION_SYSTEM_ENABLED:
            base_damage  = config.PROGRESSION_MELEE_BASE_DAMAGE
            level_damage = config.PROGRESSION_MELEE_LEVEL_DAMAGE

        offense_fn   = lambda e: e.melee_attack
        defense_fn   = lambda e: e.melee_defense
    elif skill_type == Skill.Range:
        base_damage  = config.COMBAT_RANGE_DAMAGE

        if config.PROGRESSION_SYSTEM_ENABLED:
            base_damage  = config.PROGRESSION_RANGE_BASE_DAMAGE
            level_damage = config.PROGRESSION_RANGE_LEVEL_DAMAGE

        offense_fn   = lambda e: e.range_attack
        defense_fn   = lambda e: e.range_defense
    elif skill_type == Skill.Mage:
        base_damage  = config.COMBAT_MAGE_DAMAGE

        if config.PROGRESSION_SYSTEM_ENABLED:
            base_damage  = config.PROGRESSION_MAGE_BASE_DAMAGE
            level_damage = config.PROGRESSION_MAGE_LEVEL_DAMAGE

        offense_fn   = lambda e: e.mage_attack
        defense_fn   = lambda e: e.mage_defense
    elif __debug__:
        assert False, 'Attack skill must be Melee, Range, or Mage'

    # Compute modifiers
    multiplier        = damage_multiplier(config, skill, target)
    skill_offense     = base_damage + level_damage * skill.level.val
    skill_defense     = config.PROGRESSION_BASE_DEFENSE  + config.PROGRESSION_LEVEL_DEFENSE*level(target.skills)
    equipment_offense = player.equipment.total(offense_fn)
    equipment_defense = target.equipment.total(defense_fn)

    # Total damage calculation
    offense = skill_offense + equipment_offense
    defense = skill_defense + equipment_defense
    damage  = config.COMBAT_DAMAGE_FORMULA(offense, defense, multiplier)
    #damage  = multiplier * (offense - defense)
    damage  = max(int(damage), 0)

    if config.LOG_MILESTONES and player.isPlayer and realm.quill.milestone.log_max(f'Damage_{skill_name}', damage) and config.LOG_VERBOSE:
        player_ilvl = player.equipment.total(lambda e: e.level)
        target_ilvl = target.equipment.total(lambda e: e.level)

        logging.info(f'COMBAT: Inflicted {damage} {skill_name} damage (lvl {player_level} i{player_ilvl} vs lvl {target_level} i{target_ilvl})')

    player.applyDamage(damage, skill.__class__.__name__.lower())
    target.receiveDamage(player, damage)

    return damage

def danger(config, pos, full=False):
   border = config.MAP_BORDER
   center = config.MAP_CENTER
   r, c   = pos
  
   #Distance from border
   rDist  = min(r - border, center + border - r - 1)
   cDist  = min(c - border, center + border - c - 1)
   dist   = min(rDist, cDist)
   norm   = 2 * dist / center

   if full:
      return norm, mag

   return norm

def spawn(config, dnger):
    border = config.MAP_BORDER
    center = config.MAP_CENTER
    mid    = center // 2

    dist       = dnger * center / 2
    max_offset = mid - dist
    offset     = mid + border + np.random.randint(-max_offset, max_offset)

    rng = np.random.rand()
    if rng < 0.25:
        r = border + dist
        c = offset
    elif rng < 0.5:
        r = border + center - dist - 1
        c = offset
    elif rng < 0.75:
        c = border + dist
        r = offset
    else:
        c = border + center - dist - 1
        r = offset

    if __debug__:
        assert dnger == danger(config, (r,c)), 'Agent spawned at incorrect radius'

    r = int(r)
    c = int(c)

    return r, c



from pdb import set_trace as T

from typing import Callable
from dataclasses import dataclass


@dataclass
class Task:
    condition: Callable
    target: float = None
    reward: float = 0


class Diary:
   def __init__(self, tasks):
      self.achievements = []
      for task in tasks:
         self.achievements.append(Achievement(task.condition, task.target, task.reward))

   @property
   def completed(self):
      return sum(a.completed for a in self.achievements)

   @property
   def cumulative_reward(self, aggregate=True):
      return sum(a.reward * a.completed for a in self.achievements)

   def update(self, realm, entity):
      return {a.name: a.update(realm, entity) for a in self.achievements}


class Achievement:
   def __init__(self, condition, target, reward):
      self.completed = False

      self.condition = condition
      self.target    = target
      self.reward    = reward

   @property
   def name(self):
      return '{}_{}'.format(self.condition.__name__, self.target)

   def update(self, realm, entity):
      if self.completed:
         return 0

      metric = self.condition(realm, entity)

      if metric >= self.target:
         self.completed = True
         return self.reward

      return 0



from pdb import set_trace as T
import numpy as np

from enum import Enum, auto

import nmmo
from nmmo.lib import utils
from nmmo.lib.utils import staticproperty

class NodeType(Enum):
   #Tree edges
   STATIC = auto()    #Traverses all edges without decisions
   SELECTION = auto() #Picks an edge to follow

   #Executable actions
   ACTION    = auto() #No arguments
   CONSTANT  = auto() #Constant argument
   VARIABLE  = auto() #Variable argument

class Node(metaclass=utils.IterableNameComparable):
   @classmethod
   def init(cls, config):
       pass

   @staticproperty
   def edges():
      return []

   #Fill these in
   @staticproperty
   def priority():
      return None

   @staticproperty
   def type():
      return None

   @staticproperty
   def leaf():
      return False

   @classmethod
   def N(cls, config):
      return len(cls.edges)

   def deserialize(realm, entity, index):
      return index

   def args(stim, entity, config):
      return []

class Fixed:
   pass

#ActionRoot
class Action(Node):
   nodeType = NodeType.SELECTION
   hooked   = False

   @classmethod
   def init(cls, config):
      # Sets up serialization domain
      if Action.hooked:
          return

      Action.hooked = True

   #Called upon module import (see bottom of file)
   #Sets up serialization domain
   def hook(config):
      idx = 0
      arguments = []
      for action in Action.edges(config):
         action.init(config)
         for args in action.edges:
            args.init(config)
            if not 'edges' in args.__dict__:
               continue
            for arg in args.edges: 
               arguments.append(arg)
               arg.serial = tuple([idx])
               arg.idx = idx 
               idx += 1
      Action.arguments = arguments

   @staticproperty
   def n():
      return len(Action.arguments)

   @classmethod
   def edges(cls, config):
      '''List of valid actions'''
      edges = [Move]
      if config.COMBAT_SYSTEM_ENABLED:
          edges.append(Attack)
      if config.ITEM_SYSTEM_ENABLED:
          edges += [Use]
      if config.EXCHANGE_SYSTEM_ENABLED:
          edges += [Buy, Sell]
      if config.COMMUNICATION_SYSTEM_ENABLED:
          edges.append(Comm)
      return edges

   def args(stim, entity, config):
      return nmmo.Serialized.edges 

class Move(Node):
   priority = 1
   nodeType = NodeType.SELECTION
   def call(env, entity, direction):
      r, c  = entity.pos
      entID = entity.entID
      entity.history.lastPos = (r, c)
      rDelta, cDelta = direction.delta
      rNew, cNew = r+rDelta, c+cDelta
      
      #One agent per cell
      tile = env.map.tiles[rNew, cNew] 
      if tile.occupied and not tile.lava:
         return

      if entity.status.freeze > 0:
         return

      env.dataframe.move(nmmo.Serialized.Entity, entID, (r, c), (rNew, cNew))
      entity.base.r.update(rNew)
      entity.base.c.update(cNew)

      env.map.tiles[r, c].delEnt(entID)
      env.map.tiles[rNew, cNew].addEnt(entity)

      if env.map.tiles[rNew, cNew].lava:
         entity.receiveDamage(None, entity.resources.health.val)

   @staticproperty
   def edges():
      return [Direction]

   @staticproperty
   def leaf():
      return True

class Direction(Node):
   argType = Fixed

   @staticproperty
   def edges():
      return [North, South, East, West]

   def args(stim, entity, config):
      return Direction.edges

class North(Node):
   delta = (-1, 0)

class South(Node):
   delta = (1, 0)

class East(Node):
   delta = (0, 1)

class West(Node):
   delta = (0, -1)


class Attack(Node):
   priority = 0
   nodeType = NodeType.SELECTION
   @staticproperty
   def n():
      return 3

   @staticproperty
   def edges():
      return [Style, Target]

   @staticproperty
   def leaf():
      return True

   def inRange(entity, stim, config, N):
      R, C = stim.shape
      R, C = R//2, C//2

      rets = set([entity])
      for r in range(R-N, R+N+1):
         for c in range(C-N, C+N+1):
            for e in stim[r, c].ents.values():
               rets.add(e)

      rets = list(rets)
      return rets

   def l1(pos, cent):
      r, c = pos
      rCent, cCent = cent
      return abs(r - rCent) + abs(c - cCent)

   def call(env, entity, style, targ):
      config = env.config

      if entity.isPlayer and not config.COMBAT_SYSTEM_ENABLED:
         return 

      # Testing a spawn immunity against old agents to avoid spawn camping
      immunity = config.COMBAT_SPAWN_IMMUNITY
      if entity.isPlayer and targ.isPlayer and entity.history.timeAlive.val > immunity and targ.history.timeAlive < immunity:
         return

      #Check if self targeted
      if entity.entID == targ.entID:
         return

      #ADDED: POPULATION IMMUNITY
      if not config.COMBAT_FRIENDLY_FIRE and entity.isPlayer and entity.base.population.val == targ.base.population.val:
         return

      #Check attack range
      rng     = style.attackRange(config)
      start   = np.array(entity.base.pos)
      end     = np.array(targ.base.pos)
      dif     = np.max(np.abs(start - end))

      #Can't attack same cell or out of range
      if dif == 0 or dif > rng:
         return 
      
      #Execute attack
      entity.history.attack = {}
      entity.history.attack['target'] = targ.entID
      entity.history.attack['style'] = style.__name__
      targ.attacker = entity
      targ.attackerID.update(entity.entID)

      from nmmo.systems import combat
      dmg = combat.attack(env, entity, targ, style.skill)

      if style.freeze and dmg > 0:
         targ.status.freeze.update(config.COMBAT_FREEZE_TIME)

      return dmg

class Style(Node):
   argType = Fixed
   @staticproperty
   def edges():
      return [Melee, Range, Mage]

   def args(stim, entity, config):
      return Style.edges


class Target(Node):
   argType = None

   @classmethod
   def N(cls, config):
      #return config.WINDOW ** 2
      return config.PLAYER_N_OBS

   def deserialize(realm, entity, index):
      return realm.entity(index)

   def args(stim, entity, config):
      #Should pass max range?
      return Attack.inRange(entity, stim, config, None)

class Melee(Node):
   nodeType = NodeType.ACTION
   freeze=False

   def attackRange(config):
      return config.COMBAT_MELEE_REACH

   def skill(entity):
      return entity.skills.melee

class Range(Node):
   nodeType = NodeType.ACTION
   freeze=False

   def attackRange(config):
      return config.COMBAT_RANGE_REACH

   def skill(entity):
      return entity.skills.range

class Mage(Node):
   nodeType = NodeType.ACTION
   freeze=False

   def attackRange(config):
      return config.COMBAT_MAGE_REACH

   def skill(entity):
      return entity.skills.mage

class Use(Node):
    priority = 3

    @staticproperty
    def edges():
        return [Item]

    def call(env, entity, item):
        if item not in entity.inventory:
            return

        return item.use(entity)

class Give(Node):
    priority = 2

    @staticproperty
    def edges():
        return [Item, Target]

    def call(env, entity, item, target):
        if item not in entity.inventory:
            return

        if not target.isPlayer:
            return

        if not target.inventory.space:
            return

        entity.inventory.remove(item, quantity=1)
        item = type(item)(env, item.level.val)
        target.inventory.receive(item)

        return True


class Item(Node):
    argType  = 'Entity'

    @classmethod
    def N(cls, config):
        return config.ITEM_N_OBS

    def args(stim, entity, config):
        return stim.exchange.items()

    def deserialize(realm, entity, index):
        return realm.items[index]

class Buy(Node):
    priority = 4
    argType  = Fixed

    @staticproperty
    def edges():
        return [Item]

    def call(env, entity, item):
        #Do not process exchange actions on death tick
        if not entity.alive:
            return

        if not entity.inventory.space:
            return

        return env.exchange.buy(env, entity, item)

class Sell(Node):
    priority = 4
    argType  = Fixed

    @staticproperty
    def edges():
        return [Item, Price]

    def call(env, entity, item, price):
        #Do not process exchange actions on death tick
        if not entity.alive:
            return

        # TODO: Find a better way to check this
        # Should only occur when item is used on same tick
        # Otherwise should not be possible
        if item not in entity.inventory:
            return

        if type(price) != int:
            price = price.val

        return env.exchange.sell(env, entity, item, price)

def init_discrete(values):
    classes = []
    for i in values:
        name = f'Discrete_{i}'
        cls  = type(name, (object,), {'val': i})
        classes.append(cls)
    return classes

class Price(Node):
    argType  = Fixed

    @classmethod
    def init(cls, config):
        Price.classes = init_discrete(list(range(100)))

    @staticproperty
    def edges():
        return Price.classes

    def args(stim, entity, config):
        return Price.edges

class Token(Node):
    argType  = Fixed

    @classmethod
    def init(cls, config):
        Comm.classes = init_discrete(range(config.COMMUNICATION_NUM_TOKENS))

    @staticproperty
    def edges():
        return Comm.classes

    def args(stim, entity, config):
        return Comm.edges

class Comm(Node):
    argType  = Fixed
    priority = 0

    @staticproperty
    def edges():
        return [Token]

    def call(env, entity, token):
        entity.base.comm.update(token.val)

#TODO: Solve AGI
class BecomeSkynet:
   pass



from pdb import set_trace as T

import logging
import random

from nmmo.io.stimulus import Serialized
from nmmo.lib.colors import Tier
from nmmo.systems import combat

class ItemID:
   item_ids = {} 
   id_items = {}

   def register(cls, item_id):
      if __debug__:
         if cls in ItemID.item_ids:
            assert ItemID.item_ids[cls] == item_id, f'Missmatched item_id assignment for class {cls}'
         if item_id in ItemID.id_items:
            assert ItemID.id_items[item_id] == cls, f'Missmatched class assignment for item_id {item_id}'

      ItemID.item_ids[cls] = item_id
      ItemID.id_items[item_id] = cls

   def get(cls_or_id):
      if type(cls_or_id) == int:
         return ItemID.id_items[cls_or_id]
      return ItemID.item_ids[cls_or_id]

class Item:
   ITEM_ID = None
   INSTANCE_ID = 0
   def __init__(self, realm, level,
         capacity=0, quantity=1, tradable=True,
         melee_attack=0, range_attack=0, mage_attack=0,
         melee_defense=0, range_defense=0, mage_defense=0,
         health_restore=0, resource_restore=0, price=0):

      self.config     = realm.config
      self.realm      = realm  

      self.instanceID = Item.INSTANCE_ID
      realm.items[self.instanceID] = self

      self.instance         = Serialized.Item.ID(realm.dataframe, self.instanceID, Item.INSTANCE_ID)
      self.index            = Serialized.Item.Index(realm.dataframe, self.instanceID, self.ITEM_ID)
      self.level            = Serialized.Item.Level(realm.dataframe, self.instanceID, level)
      self.capacity         = Serialized.Item.Capacity(realm.dataframe, self.instanceID, capacity)
      self.quantity         = Serialized.Item.Quantity(realm.dataframe, self.instanceID, quantity)
      self.tradable         = Serialized.Item.Tradable(realm.dataframe, self.instanceID, tradable)
      self.melee_attack     = Serialized.Item.MeleeAttack(realm.dataframe, self.instanceID, melee_attack)
      self.range_attack     = Serialized.Item.RangeAttack(realm.dataframe, self.instanceID, range_attack)
      self.mage_attack      = Serialized.Item.MageAttack(realm.dataframe, self.instanceID, mage_attack)
      self.melee_defense    = Serialized.Item.MeleeDefense(realm.dataframe, self.instanceID, melee_defense)
      self.range_defense    = Serialized.Item.RangeDefense(realm.dataframe, self.instanceID, range_defense)
      self.mage_defense     = Serialized.Item.MageDefense(realm.dataframe, self.instanceID, mage_defense)
      self.health_restore   = Serialized.Item.HealthRestore(realm.dataframe, self.instanceID, health_restore)
      self.resource_restore = Serialized.Item.ResourceRestore(realm.dataframe, self.instanceID, resource_restore)
      self.price            = Serialized.Item.Price(realm.dataframe, self.instanceID, price)
      self.equipped         = Serialized.Item.Equipped(realm.dataframe, self.instanceID, 0)

      realm.dataframe.init(Serialized.Item, self.instanceID, None)

      Item.INSTANCE_ID += 1
      if self.ITEM_ID is not None:
         ItemID.register(self.__class__, item_id=self.ITEM_ID)

   @property
   def signature(self):
      return (self.index.val, self.level.val)

   @property
   def packet(self):
      return {'item':             self.__class__.__name__,
              'level':            self.level.val,
              'capacity':         self.capacity.val,
              'quantity':         self.quantity.val,
              'melee_attack':     self.melee_attack.val,
              'range_attack':     self.range_attack.val,
              'mage_attack':      self.mage_attack.val,
              'melee_defense':    self.melee_defense.val,
              'range_defense':    self.range_defense.val,
              'mage_defense':     self.mage_defense.val,
              'health_restore':   self.health_restore.val,
              'resource_restore': self.resource_restore.val,
              'price':            self.price.val}
 
   def use(self, entity):
      return
      #TODO: Warning?
      #assert False, f'Use {type(self)} not defined'

class Stack():
    pass

class Gold(Item, Stack):
   ITEM_ID = 1
   def __init__(self, realm, **kwargs):
      super().__init__(realm, level=0, tradable=False, **kwargs)

class Equipment(Item):
   @property
   def packet(self):
     packet = {'color': self.color.packet()}
     return {**packet, **super().packet}

   @property
   def color(self):
     if self.level == 0:
        return Tier.BLACK
     if self.level < 10:
        return Tier.WOOD
     elif self.level < 20:
        return Tier.BRONZE
     elif self.level < 40:
        return Tier.SILVER
     elif self.level < 60:
        return Tier.GOLD
     elif self.level < 80:
        return Tier.PLATINUM
     else:
        return Tier.DIAMOND

   def use(self, entity):
      if self.equipped.val:
         self.equipped.update(0)
         equip = self.unequip(entity)
      else:
         self.equipped.update(1)
         equip = self.equip(entity)

         config = self.config
         if not config.LOG_MILESTONES or not entity.isPlayer:
             return equip

         realm     = self.realm
         equipment = entity.equipment
         item_name = self.__class__.__name__

         if realm.quill.milestone.log_max(f'{item_name}_level', self.level.val) and config.LOG_VERBOSE:
            logging.info(f'EQUIPMENT: Equipped level {self.level.val} {item_name}')
         if realm.quill.milestone.log_max(f'Item_Level', equipment.item_level) and config.LOG_VERBOSE:
            logging.info(f'EQUIPMENT: Item level {equipment.item_level}')
         if realm.quill.milestone.log_max(f'Mage_Attack', equipment.mage_attack) and config.LOG_VERBOSE:
            logging.info(f'EQUIPMENT: Mage attack {equipment.mage_attack}')
         if realm.quill.milestone.log_max(f'Mage_Defense', equipment.mage_defense) and config.LOG_VERBOSE:
            logging.info(f'EQUIPMENT: Mage defense {equipment.mage_defense}')
         if realm.quill.milestone.log_max(f'Range_Attack', equipment.range_attack) and config.LOG_VERBOSE:
            logging.info(f'EQUIPMENT: Range attack {equipment.range_attack}')
         if realm.quill.milestone.log_max(f'Range_Defense', equipment.range_defense) and config.LOG_VERBOSE:
            logging.info(f'EQUIPMENT: Range defense {equipment.range_defense}')
         if realm.quill.milestone.log_max(f'Melee_Attack', equipment.melee_attack) and config.LOG_VERBOSE:
            logging.info(f'EQUIPMENT: Melee attack {equipment.melee_attack}')
         if realm.quill.milestone.log_max(f'Melee_Defense', equipment.melee_defense) and config.LOG_VERBOSE:
                logging.info(f'EQUIPMENT: Melee defense {equipment.melee_defense}')

      return equip
 
class Armor(Equipment):
   def __init__(self, realm, level, **kwargs):
      defense = realm.config.EQUIPMENT_ARMOR_BASE_DEFENSE + level*realm.config.EQUIPMENT_ARMOR_LEVEL_DEFENSE
      super().__init__(realm, level,
              melee_defense=defense,
              range_defense=defense,
              mage_defense=defense,
              **kwargs)

class Hat(Armor):
   ITEM_ID = 2

   def equip(self, entity):
      if entity.level < self.level.val:
          return
      if entity.inventory.equipment.hat:
          entity.inventory.equipment.hat.use(entity)
      entity.inventory.equipment.hat = self

   def unequip(self, entity):
      entity.inventory.equipment.hat = None

class Top(Armor):
   ITEM_ID = 3

   def equip(self, entity):
      if entity.level < self.level.val:
          return
      if entity.inventory.equipment.top:
          entity.inventory.equipment.top.use(entity)
      entity.inventory.equipment.top = self

   def unequip(self, entity):
      entity.inventory.equipment.top = None

class Bottom(Armor):
   ITEM_ID = 4

   def equip(self, entity):
      if entity.level < self.level.val:
          return
      if entity.inventory.equipment.bottom:
          entity.inventory.equipment.bottom.use(entity)
      entity.inventory.equipment.bottom = self

   def unequip(self, entity):
      entity.inventory.equipment.bottom = None

class Weapon(Equipment):
   def __init__(self, realm, level, **kwargs):
      super().__init__(realm, level, **kwargs)
      self.attack = realm.config.EQUIPMENT_WEAPON_BASE_DAMAGE + level*realm.config.EQUIPMENT_WEAPON_LEVEL_DAMAGE

   def equip(self, entity):
      if entity.inventory.equipment.held:
          entity.inventory.equipment.held.use(entity)
      entity.inventory.equipment.held = self

   def unequip(self, entity):
      entity.inventory.equipment.held = None

class Sword(Weapon):
   ITEM_ID = 5
   def __init__(self, realm, level, **kwargs):
      super().__init__(realm, level, **kwargs)
      self.melee_attack.update(self.attack)

   def equip(self, entity):
      if entity.skills.melee.level.val >= self.level.val:
         super().equip(entity)
 
class Bow(Weapon):
   ITEM_ID = 6
   def __init__(self, realm, level, **kwargs):
      super().__init__(realm, level, **kwargs)
      self.range_attack.update(self.attack)

   def equip(self, entity):
      if entity.skills.range.level.val >= self.level.val:
         super().equip(entity)
 
class Wand(Weapon):
   ITEM_ID = 7
   def __init__(self, realm, level, **kwargs):
      super().__init__(realm, level, **kwargs)
      self.mage_attack.update(self.attack)

   def equip(self, entity):
      if entity.skills.mage.level.val >= self.level.val:
         super().equip(entity)
 
class Tool(Equipment):
   def __init__(self, realm, level, **kwargs):
      defense = realm.config.EQUIPMENT_TOOL_BASE_DEFENSE + level*realm.config.EQUIPMENT_TOOL_LEVEL_DEFENSE
      super().__init__(realm, level,
              melee_defense=defense,
              range_defense=defense,
              mage_defense=defense,
              **kwargs)

   def equip(self, entity):
      if entity.inventory.equipment.held:
          entity.inventory.equipment.held.use(entity)
      entity.inventory.equipment.held = self

   def unequip(self, entity):
      entity.inventory.equipment.held = None

class Rod(Tool):
    ITEM_ID = 8
    def equip(self, entity):
       if entity.skills.fishing.level >= self.level.val:
          super().equip(entity)
          return True

       return False

class Gloves(Tool):
    ITEM_ID = 9
    def equip(self, entity):
       if entity.skills.herbalism.level >= self.level.val:
          super().equip(entity)
          return True

       return False

class Pickaxe(Tool):
    ITEM_ID = 10
    def equip(self, entity):
       if entity.skills.prospecting.level >= self.level.val:
          super().equip(entity)
          return True

       return False

class Chisel(Tool):
    ITEM_ID = 11
    def equip(self, entity):
       if entity.skills.carving.level >= self.level.val:
          super().equip(entity)
          return True

       return False

class Arcane(Tool):
    ITEM_ID = 12
    def equip(self, entity):
       if entity.skills.alchemy.level >= self.level.val:
          super().equip(entity)
          return True

       return False

class Ammunition(Equipment, Stack):
   def __init__(self, realm, level, **kwargs):
       super().__init__(realm, level, **kwargs)
       self.attack = realm.config.EQUIPMENT_AMMUNITION_BASE_DAMAGE + level*realm.config.EQUIPMENT_AMMUNITION_LEVEL_DAMAGE

   def equip(self, entity):
      if entity.inventory.equipment.ammunition:
          entity.inventory.equipment.ammunition.use(entity)
      entity.inventory.equipment.ammunition = self

   def unequip(self, entity):
      entity.inventory.equipment.ammunition = None

   def fire(self, entity):
      if __debug__:
         err = 'Used ammunition with 0 quantity'
         assert self.quantity.val > 0, err

      self.quantity.decrement()

      if self.quantity.val == 0:
         entity.inventory.remove(self)

      return self.damage
      
class Scrap(Ammunition):
   ITEM_ID = 13
   def __init__(self, realm, level, **kwargs):
      super().__init__(realm, level, **kwargs)
      self.melee_attack.update(self.attack)

   def equip(self, entity):
      if entity.skills.melee.level >= self.level.val:
          super().equip(entity)
          return True

      return False

   @property
   def damage(self):
      return self.melee_attack.val

class Shaving(Ammunition):
   ITEM_ID = 14
   def __init__(self, realm, level, **kwargs):
      super().__init__(realm, level, **kwargs)
      self.range_attack.update(self.attack)

   def equip(self, entity):
      if entity.skills.range.level >= self.level.val:
          super().equip(entity)
          return True

      return False

   @property
   def damage(self):
      return self.range_attack.val

class Shard(Ammunition):
   ITEM_ID = 15
   def __init__(self, realm, level, **kwargs):
      super().__init__(realm, level, **kwargs)
      self.mage_attack.update(self.attack)

   def equip(self, entity):
      if entity.skills.mage.level >= self.level.val:
          super().equip(entity)
          return True

      return False

   @property
   def damage(self):
      return self.mage_attack.val

class Consumable(Item):
    pass

class Ration(Consumable):
   ITEM_ID = 16
   def __init__(self, realm, level, **kwargs):
      restore = realm.config.PROFESSION_CONSUMABLE_RESTORE(level)
      super().__init__(realm, level, resource_restore=restore, **kwargs)

   def use(self, entity):
      if entity.level < self.level.val:
          return False

      if self.config.LOG_MILESTONES and self.realm.quill.milestone.log_max(f'Consumed_Ration', self.level.val) and self.config.LOG_VERBOSE:
         logging.info(f'PROFESSION: Consumed level {self.level.val} ration')

      entity.resources.food.increment(self.resource_restore.val)
      entity.resources.water.increment(self.resource_restore.val)

      entity.ration_level_consumed = max(entity.ration_level_consumed, self.level.val)
      entity.ration_consumed += 1

      entity.inventory.remove(self)

      return True

class Poultice(Consumable):
   ITEM_ID = 17

   def __init__(self, realm, level, **kwargs):
      restore = realm.config.PROFESSION_CONSUMABLE_RESTORE(level)
      super().__init__(realm, level, health_restore=restore, **kwargs)

   def use(self, entity):
      if entity.level < self.level.val:
          return False

      if self.config.LOG_MILESTONES and self.realm.quill.milestone.log_max(f'Consumed_Poultice', self.level.val) and self.config.LOG_VERBOSE:
         logging.info(f'PROFESSION: Consumed level {self.level.val} poultice')

      entity.resources.health.increment(self.health_restore.val)

      entity.poultice_level_consumed = max(entity.poultice_level_consumed, self.level.val)
      entity.poultice_consumed       += 1

      entity.inventory.remove(self)

      return True



from pdb import set_trace as T
import numpy as np

from nmmo.lib import utils

class SerializedVariable:
   CONTINUOUS = False
   DISCRETE   = False
   def __init__(self, dataframe, key, val=None, config=None):
      if config is None:
         config    = dataframe.config
 
      self.obj  = str(self.__class__).split('.')[-2]
      self.attr = self.__class__.__name__
      self.key  = key

      self.min = 0
      self.max = np.inf
      self.val = val

      self.dataframe = dataframe
      self.init(config)
      err = 'Must set a default val upon instantiation or init()'
      assert self.val is not None, err

      #Update dataframe
      if dataframe is not None:
         self.update(self.val)

   #Defined for cleaner stim files
   def init(self):
      pass

   def packet(self):
      return {
            'val': self.val,
            'max': self.max}

   def update(self, val):
      self.val = min(max(val, self.min), self.max)
      self.dataframe.update(self, self.val)
      return self

   def increment(self, val=1):
      self.update(self.val + val)
      return self

   def decrement(self, val=1):
      self.update(self.val - val)
      return self

   @property
   def empty(self):
      return self.val == 0

   def __add__(self, other):
      self.increment(other)
      return self

   def __sub__(self, other):
      self.decrement(other)
      return self

   def __eq__(self, other):
      return self.val == other

   def __ne__(self, other):
      return self.val != other

   def __lt__(self, other):
      return self.val < other

   def __le__(self, other):
      return self.val <= other

   def __gt__(self, other):
      return self.val > other

   def __ge__(self, other):
      return self.val >= other

class Continuous(SerializedVariable):
   CONTINUOUS = True

class Discrete(Continuous):
   DISCRETE = True


class Serialized(metaclass=utils.IterableNameComparable):
   def dict():
      return {k[0] : v for k, v in dict(Stimulus).items()}

   class Entity(metaclass=utils.IterableNameComparable):
      @staticmethod
      def enabled(config):
         return True

      @staticmethod
      def N(config):
         return config.PLAYER_N_OBS

      class Self(Discrete):
         def init(self, config):
            self.max = 1
            self.scale = 1.0

      class ID(Continuous):
         def init(self, config):
            self.min   = -np.inf
            self.scale = 0.001

      class AttackerID(Continuous):
         def init(self, config):
            self.min   = -np.inf
            self.scale = 0.001

      class Level(Continuous):
         def init(self, config):
            self.scale = 0.05

      class ItemLevel(Continuous):
         def init(self, config):
            self.scale = 0.025
            self.max   = 5 * config.NPC_LEVEL_MAX

      class Comm(Discrete):
         def init(self, config):
            self.scale = 0.025
            self.max = 1
            if config.COMMUNICATION_SYSTEM_ENABLED:
                self.max   = config.COMMUNICATION_NUM_TOKENS

      class Population(Discrete):
         def init(self, config):
            self.min = -3 #NPC index
            self.max = config.PLAYER_POLICIES - 1
            self.scale = 1.0

      class R(Discrete):
         def init(self, config):
            self.min = 0
            self.max = config.MAP_SIZE - 1
            self.scale = 0.15

      class C(Discrete):
         def init(self, config):
            self.min = 0
            self.max = config.MAP_SIZE - 1
            self.scale = 0.15

      # Historical stats
      class Damage(Continuous):
         def init(self, config):
            #This scale may eventually be too high
            self.val   = 0
            self.scale = 0.1

      class TimeAlive(Continuous):
         def init(self, config):
            self.val = 0
            self.scale = 0.01

      # Status effects
      class Freeze(Continuous):
         def init(self, config):
            self.val = 0
            self.max = 3
            self.scale = 0.3

      class Gold(Continuous):
         def init(self, config):
            self.val = 0
            self.scale = 0.01

      # Resources -- Redo the max/min scaling. You can't change these
      # after init without messing up the embeddings
      class Health(Continuous):
         def init(self, config):
            self.val = config.PLAYER_BASE_HEALTH
            self.max = config.PLAYER_BASE_HEALTH
            self.scale = 0.1

      class Food(Continuous):
         def init(self, config):
            if config.RESOURCE_SYSTEM_ENABLED:
               self.val = config.RESOURCE_BASE
               self.max = config.RESOURCE_BASE
            else:
               self.val = 1
               self.max = 1
    
            self.scale = 0.01

      class Water(Continuous):
         def init(self, config):
            if config.RESOURCE_SYSTEM_ENABLED:
               self.val = config.RESOURCE_BASE
               self.max = config.RESOURCE_BASE
            else:
               self.val = 1
               self.max = 1
 
            self.scale = 0.01

      class Melee(Continuous):
         def init(self, config):
            self.val = 1
            self.max = 1
            if config.PROGRESSION_SYSTEM_ENABLED:
                self.max = config.PROGRESSION_LEVEL_MAX

      class Range(Continuous):
         def init(self, config):
            self.val = 1
            self.max = 1
            if config.PROGRESSION_SYSTEM_ENABLED:
                self.max = config.PROGRESSION_LEVEL_MAX

      class Mage(Continuous):
         def init(self, config):
            self.val = 1
            self.max = 1
            if config.PROGRESSION_SYSTEM_ENABLED:
                self.max = config.PROGRESSION_LEVEL_MAX

      class Fishing(Continuous):
         def init(self, config):
            self.val = 1
            self.max = 1
            if config.PROGRESSION_SYSTEM_ENABLED:
                self.max = config.PROGRESSION_LEVEL_MAX

      class Herbalism(Continuous):
         def init(self, config):
            self.val = 1
            self.max = 1
            if config.PROGRESSION_SYSTEM_ENABLED:
                self.max = config.PROGRESSION_LEVEL_MAX

      class Prospecting(Continuous):
         def init(self, config):
            self.val = 1
            self.max = 1
            if config.PROGRESSION_SYSTEM_ENABLED:
                self.max = config.PROGRESSION_LEVEL_MAX

      class Carving(Continuous):
         def init(self, config):
            self.val = 1
            self.max = 1
            if config.PROGRESSION_SYSTEM_ENABLED:
                self.max = config.PROGRESSION_LEVEL_MAX

      class Alchemy(Continuous):
         def init(self, config):
            self.val = 1
            self.max = 1
            if config.PROGRESSION_SYSTEM_ENABLED:
                self.max = config.PROGRESSION_LEVEL_MAX

   class Tile(metaclass=utils.IterableNameComparable):
      @staticmethod
      def enabled(config):
         return True

      @staticmethod
      def N(config):
         return config.MAP_N_OBS

      class NEnts(Continuous):
         def init(self, config):
            self.max = config.PLAYER_N
            self.val = 0
            self.scale = 1.0

      class Index(Discrete):
         def init(self, config):
            self.max = config.MAP_N_TILE
            self.scale = 0.15

      class R(Discrete):
         def init(self, config):
            self.max = config.MAP_SIZE - 1
            self.scale = 0.15
 
      class C(Discrete):
         def init(self, config):
            self.max = config.MAP_SIZE - 1
            self.scale = 0.15

   class Item(metaclass=utils.IterableNameComparable):
      @staticmethod
      def enabled(config):
         return config.ITEM_SYSTEM_ENABLED

      @staticmethod
      def N(config):
         return config.ITEM_N_OBS

      class ID(Continuous):
         def init(self, config):
            self.scale = 0.001

      class Index(Discrete):
         def init(self, config):
            self.max   = config.ITEM_N + 1
            self.scale = 1.0 / self.max

      class Level(Continuous):
         def init(self, config):
            self.max   = 99
            self.scale = 1.0 / self.max

      class Capacity(Continuous):
         def init(self, config):
            self.max   = 99
            self.scale = 1.0 / self.max

      class Quantity(Continuous):
         def init(self, config):
            self.max   = 99
            self.scale = 1.0 / self.max

      class Tradable(Discrete):
         def init(self, config):
            self.max   = 1
            self.scale = 1.0

      class MeleeAttack(Continuous):
         def init(self, config):
            self.max   = 100
            self.scale = 1.0 / self.max

      class RangeAttack(Continuous):
         def init(self, config):
            self.max   = 100
            self.scale = 1.0 / self.max

      class MageAttack(Continuous):
         def init(self, config):
            self.max   = 100
            self.scale = 1.0 / self.max

      class MeleeDefense(Continuous):
         def init(self, config):
            self.max   = 100
            self.scale = 1.0 / self.max

      class RangeDefense(Continuous):
         def init(self, config):
            self.max   = 100
            self.scale = 1.0 / self.max

      class MageDefense(Continuous):
         def init(self, config):
            self.max   = 100
            self.scale = 1.0 / self.max

      class HealthRestore(Continuous):
         def init(self, config):
            self.max   = 100
            self.scale = 1.0 / self.max

      class ResourceRestore(Continuous):
         def init(self, config):
            self.max   = 100
            self.scale = 1.0 / self.max

      class Price(Continuous):
         def init(self, config):
            self.scale = 0.01

      class Equipped(Discrete):
         def init(self, config):
            self.scale = 1.0

   # TODO: Figure out how to autogen this from Items
   class Market(metaclass=utils.IterableNameComparable):
      @staticmethod
      def enabled(config):
         return config.EXCHANGE_SYSTEM_ENABLED

      @staticmethod
      def N(config):
         return config.EXCHANGE_N_OBS

      class ID(Continuous):
         def init(self, config):
            self.scale = 0.001

      class Index(Discrete):
         def init(self, config):
            self.max   = config.ITEM_N + 1
            self.scale = 1.0 / self.max

      class Level(Continuous):
         def init(self, config):
            self.max   = 99
            self.scale = 1.0 / self.max

      class Capacity(Continuous):
         def init(self, config):
            self.max   = 99
            self.scale = 1.0 / self.max

      class Quantity(Continuous):
         def init(self, config):
            self.max   = 99
            self.scale = 1.0 / self.max

      class Tradable(Discrete):
         def init(self, config):
            self.max   = 1
            self.scale = 1.0

      class MeleeAttack(Continuous):
         def init(self, config):
            self.max   = 100
            self.scale = 1.0 / self.max

      class RangeAttack(Continuous):
         def init(self, config):
            self.max   = 100
            self.scale = 1.0 / self.max

      class MageAttack(Continuous):
         def init(self, config):
            self.max   = 100
            self.scale = 1.0 / self.max

      class MeleeDefense(Continuous):
         def init(self, config):
            self.max   = 100
            self.scale = 1.0 / self.max

      class RangeDefense(Continuous):
         def init(self, config):
            self.max   = 100
            self.scale = 1.0 / self.max

      class MageDefense(Continuous):
         def init(self, config):
            self.max   = 100
            self.scale = 1.0 / self.max

      class HealthRestore(Continuous):
         def init(self, config):
            self.max   = 100
            self.scale = 1.0 / self.max

      class ResourceRestore(Continuous):
         def init(self, config):
            self.max   = 100
            self.scale = 1.0 / self.max

      class Price(Continuous):
         def init(self, config):
            self.scale = 0.01

      class Equipped(Discrete):
         def init(self, config):
            self.scale = 1.0


for objName, obj in Serialized:
   for idx, (attrName, attr) in enumerate(obj):
      attr.index = idx 



from pdb import set_trace as T
import numpy as np

class ExperienceCalculator:
   def __init__(self, num_levels=15):
      self.exp = np.array([0] + [10*2**i for i in range(num_levels)])

   def expAtLevel(self, level):
      return int(self.exp[level - 1])

   def levelAtExp(self, exp):
      if exp >= self.exp[-1]:
         return len(self.exp)
      return np.argmin(exp >= self.exp)



#Various Enums used for handling materials, entity types, etc.
#Data texture pairs are used for enums that require textures.
#These textures are filled in by the Render class at run time.

from pdb import set_trace as T
import numpy as np
import colorsys

def rgb(h):
  h = h.lstrip('#')
  return tuple(int(h[i:i+2], 16) for i in (0, 2, 4))

def rgbNorm(h):
  h = h.lstrip('#')
  return tuple(int(h[i:i+2], 16)/255.0 for i in (0, 2, 4))

def makeColor(idx, h=1, s=1, v=1):
   r, g, b = colorsys.hsv_to_rgb(h, s, v)
   rgbval = tuple(int(255*e) for e in [r, g, b])
   hexval = '%02x%02x%02x' % rgbval
   return Color(str(idx), hexval)

class Color:
    def __init__(self, name, hexVal):
        self.name = name
        self.hex = hexVal
        self.rgb = rgb(hexVal)
        self.norm = rgbNorm(hexVal)
        self.value = self.rgb #Emulate enum

    def packet(self):
        return self.hex

class Color256:
   def make256():
      parh, parv = np.meshgrid(np.linspace(0.075, 1, 16), np.linspace(0.25, 1, 16)[::-1])
      parh, parv = parh.T.ravel(), parv.T.ravel()
      idxs = np.arange(256)
      params = zip(idxs, parh, parv)
      colors = [makeColor(idx, h=h, s=1, v=v) for idx, h, v in params]
      return colors
   colors = make256()

class Color16:
   def make():
      hues   = np.linspace(0, 1, 16)
      idxs   = np.arange(256)
      params = zip(idxs, hues)
      colors = [makeColor(idx, h=h, s=1, v=1) for idx, h in params]
      return colors
   colors = make()

class Tier:
   BLACK    = Color('BLACK', '#000000')
   WOOD     = Color('WOOD', '#784d1d')
   BRONZE   = Color('BRONZE', '#db4508')
   SILVER   = Color('SILVER', '#dedede')
   GOLD     = Color('GOLD', '#ffae00')
   PLATINUM = Color('PLATINUM', '#cd75ff')
   DIAMOND  = Color('DIAMOND', '#00bbbb')
   
class Swatch:
   def colors():
      '''Return list of swatch colors'''
      return

   def rand():
      '''Return random swatch color'''
      all_colors = colors()
      randInd = np.random.randint(0, len(all_colors))
      return all_colors[randInd]


class Neon(Swatch):
   RED      = Color('RED', '#ff0000')
   ORANGE   = Color('ORANGE', '#ff8000')
   YELLOW   = Color('YELLOW', '#ffff00')

   GREEN    = Color('GREEN', '#00ff00')
   MINT     = Color('MINT', '#00ff80')
   CYAN     = Color('CYAN', '#00ffff')

   BLUE     = Color('BLUE', '#0000ff')
   PURPLE   = Color('PURPLE', '#8000ff')
   MAGENTA  = Color('MAGENTA', '#ff00ff')

   FUCHSIA  = Color('FUCHSIA', '#ff0080')
   SPRING   = Color('SPRING', '#80ff80')
   SKY      = Color('SKY', '#0080ff')
 
   WHITE    = Color('WHITE', '#ffffff')
   GRAY     = Color('GRAY', '#666666')
   BLACK    = Color('BLACK', '#000000')

   BLOOD    = Color('BLOOD', '#bb0000')
   BROWN    = Color('BROWN', '#7a3402')
   GOLD     = Color('GOLD', '#eec600')
   SILVER   = Color('SILVER', '#b8b8b8')

   TERM     = Color('TERM', '#41ff00')
   MASK     = Color('MASK', '#d67fff')

   def colors():
      return (
              Neon.CYAN, Neon.MINT, Neon.GREEN,
              Neon.BLUE, Neon.PURPLE, Neon.MAGENTA,
              Neon.FUCHSIA, Neon.SPRING, Neon.SKY,
              Neon.RED, Neon.ORANGE, Neon.YELLOW)

class Solid(Swatch):
   BLUE       = Color('BLUE', '#1f77b4')
   ORANGE     = Color('ORANGE', '#ff7f0e')
   GREEN      = Color('GREEN', '#2ca02c')

   RED        = Color('RED',  '#D62728')
   PURPLE     = Color('PURPLE', '#9467bd')
   BROWN      = Color('BROWN', '#8c564b')

   PINK       = Color('PINK', '#e377c2')
   GREY       = Color('GREY', '#7f7f7f')
   CHARTREUSE = Color('CHARTREUSE', '#bcbd22')

   SKY        = Color('SKY', '#17becf')

   def colors():
      return (
              Solid.BLUE, Solid.ORANGE, Solid.GREEN,
              Solid.RED, Solid.PURPLE, Solid.BROWN,
              Solid.PINK, Solid.CHARTREUSE, Solid.SKY,
              Solid.GREY)

class Palette:
   def __init__(self, initial_swatch=Neon):
      self.colors = {}
      for idx, color in enumerate(initial_swatch.colors()):
          self.colors[idx] = color

   def color(self, idx):
      if idx in self.colors:
           return self.colors[idx]

      color = makeColor(idx, h=np.random.rand(), s=1, v=1)
      self.colors[idx] = color
      return color
